"""

Incomplete documentation available here:

https://docs.splunk.com/Documentation/Splunk/latest/Admin/Defaultmetaconf

Specifically, attribute-level ACls aren't discussed nor is the magic "import" directive.


LEVELS:

    0 - global  (or 1 stanza="default")
    1 - conf
    2 - stanzas
    3 - attribute
"""

from __future__ import absolute_import, unicode_literals

import re
from urllib.parse import quote, unquote

from ksconf.conf.parser import GLOBAL_STANZA, parse_conf

"""

Another way to split out the hierarchy.   (Issues:   No way to warn if too many "/"s.)

d = dict(zip(("conf", "stanza", "attribute"), (unquote(p) for p in stanza_name.split("/"))))


    {"conf" :

"""


class MetaLayer:
    def __init__(self, name):
        self.name = name
        self._data = {}             # Current level data
        self._children = {}         # Named children, levels

    def resolve(self, name):
        try:
            return self._children[name]
        except KeyError:
            baby = MetaLayer(name)
            self._children[name] = baby
            return baby

    def update(self, *args, **kwargs):
        self._data.update(*args, **kwargs)

    @property
    def data(self):
        return self._data

    def walk(self, _prefix=()):
        if self._data:
            yield _prefix
        if self._children:
            for child_name, child in self._children.items():
                yield from child.walk(_prefix=_prefix + (child_name,))

    def items(self, prefix=None):
        """ Helpful when rebuilding the input file. """
        if prefix is None:
            prefix = ()
        if self._data:
            yield prefix, self._data
        if self._children:
            for child_name, child in self._children.items():
                yield from child.items(prefix=prefix + (child_name,))


class MetaData:

    regex_access = r"(?:^|\s*,\s*)(?P<action>read|write)\s*:\s*\[\s*(?P<roles>[^\]]+?)\s*\]"

    def __init__(self):
        self._meta = MetaLayer("")

    @staticmethod
    def expand_layers(layers):
        """
        :param layers: layer of stanzas, starting with the global ending with conf/stanza/attr
        :type layers: list(dict)
        :return:  Expanded layer
        :rtype: dict
        """
        # type: (list(layers)) -> dict
        exp = {}
        for layer in layers:
            if layer:
                exp.update(layer)
        return exp

    @classmethod
    def parse_meta(cls, stanza):
        """
        Split out the values of 'access' (maybe more someday)
        :param stanza: content of a meta stanza
        :return: extended meta data
        :rtype: dict
        """
        # Do we really need to do this?  .. seems safer for now
        stanza = stanza.copy()
        if "access" in stanza:
            access = stanza["access"]
            for match in re.finditer(cls.regex_access, access):
                stanza_name = "access.{}".format(match.group("action"))
                values = [role.strip() for role in match.group("roles").split(",")]
                stanza[stanza_name] = values
        return stanza

    def get_layer(self, *names):
        node = self._meta
        for name in names:
            node = node.resolve(name)
        return node

    def get(self, *names):
        node = self._meta
        layers = [node.data]
        for name in names:
            node = node.resolve(name)
            layers.append(node.data)
        d = self.expand_layers(layers)
        # Parser access:   split out 'read' vs 'write', return as list
        # d["acesss"]
        return self.parse_meta(d)

    def feed_file(self, stream):
        conf = parse_conf(stream)
        self.feed_conf(conf)

    def feed_conf(self, conf):
        for stanza_name, stanza_data in conf.items():
            if stanza_name is GLOBAL_STANZA:
                parts = []
            else:
                parts = [unquote(p) for p in stanza_name.split("/")]
                if len(parts) == 1 and parts[0] in (GLOBAL_STANZA, "", "default", "global"):
                    parts = []
            meta_layer = self.get_layer(*parts)
            meta_layer.update(stanza_data)

    def iter_raw(self):
        """ RAW """
        return self._meta.items()

    def walk(self):
        for path in self._meta.walk():
            yield (path, self.get(*path))

    def write_stream(self, stream, sort=True):
        if sort:
            # Prefix level # to list for sorting purposes
            data = [(len(parts), parts, payload) for parts, payload in self.iter_raw()]
            data.sort()
            raw = [(i[1], i[2]) for i in data]
            del data
        else:
            raw = self.iter_raw()

        for parts, payload in raw:
            stanza = "/".join(quote(p, "") for p in parts)
            stream.write("[{}]\n".format(stanza))
            for attr in sorted(payload):
                value = payload[attr]
                if attr.startswith("#"):
                    stream.write("{0}\n".format(value))
                else:
                    stream.write("{} = {}\n".format(attr, value))
            # Trailing EOL, oh well...  fix that later
            stream.write("\n")



from __future__ import absolute_import, unicode_literals

from collections import Counter, namedtuple
from shutil import which
from subprocess import PIPE, Popen, call, list2cmdline

from ksconf.compat import cache
from ksconf.util import _xargs

GIT_BIN = "git"
GitCmdOutput = namedtuple("GitCmdOutput", ["cmd", "returncode", "stdout", "stderr", "lines"])

unitesting = False


class GitNotAvailable(Exception):
    pass


def git_cmd(args, shell=False, cwd=None, capture_std=True, encoding="utf-8"):
    if isinstance(args, tuple):
        args = list(args)
    cmdline_args = [GIT_BIN] + args
    out = None
    if capture_std:
        out = PIPE
    proc = Popen(cmdline_args, stdout=out, stderr=out, shell=shell, cwd=cwd)
    (stdout, stderr) = proc.communicate()
    if hasattr(stdout, "decode"):
        stdout = stdout.decode(encoding)
        stderr = stderr.decode(encoding)
    return GitCmdOutput(cmdline_args, proc.returncode, stdout, stderr, None)


def git_cmd_iterable(args, iterable, cwd=None, cmd_len=1024):
    base_len = sum([len(s) + 1 for s in args])
    for chunk in _xargs(iterable, cmd_len - base_len):
        p = git_cmd(args + chunk, cwd=cwd)
        if p.returncode != 0:  # pragma: no cover
            raise RuntimeError("git exited with code {}.  Command: {}".format(
                p.returncode, list2cmdline(args + chunk)))


# Shave time off of unit testing; or anything that does CLI calls from the API
@cache
def git_version():
    git_path = which(GIT_BIN)
    if not git_path:
        return None
    try:
        cmd = git_cmd(["--version"])
    except GitNotAvailable:
        return None
    return {
        "version": cmd.stdout.strip(),
        "path": git_path,
    }


def git_status_summary(path):
    c = Counter()
    cmd = git_cmd(["status", "--porcelain", "--ignored", "."], cwd=path)
    if cmd.returncode != 0:  # pragma: no cover
        raise RuntimeError("git command returned exit code {}.".format(cmd.returncode))
    # XY:  X=index, Y=working tree.   For our simplistic approach we consider them together.
    for line in cmd.stdout.splitlines():
        state = line[0:2]
        if state == "??":
            c["untracked"] += 1
        elif state == "!!":
            c["ignored"] += 1
        else:
            c["changed"] += 1
    return c


'''
def get_gitdir(path=None):
    # May not need this.  the 'git status' was missing '.' to make it specific to JUST the app folder
    # I thought I needed this because of my local testing git-inside-of-git scenario...
    p = git_cmd(["rev-parse", "--git-dir"], cwd=path)
    if p.returncode == 0:
        gitdir = p.stdout.strip()
        return gitdir
    # Then later you can use  git --git-dir=apps/.git --working-tree apps Splunk_TA_aws
'''


def git_is_working_tree(path=None):
    return git_cmd(["rev-parse", "--is-inside-work-tree"], cwd=path).returncode == 0


def git_is_clean(path=None, check_untracked=True, check_ignored=False):
    # ANY change to the index or working tree is considered unclean.
    c = git_status_summary(path)
    total_changes = c["changed"]
    if check_untracked:
        total_changes += c["untracked"]
    if check_ignored:
        total_changes += c["ignored"]
    '''
    print "GIT IS CLEAN?:   path={} check_untracked={} check_ignored={} total_changes={} c={}".format(
        path, check_untracked, check_ignored, total_changes, c)
    '''
    return total_changes == 0


def git_ls_files(path, *modifiers):
    # staged=True
    args = ["ls-files"]
    for m in modifiers:
        args.append("--" + m)
    proc = git_cmd(args, cwd=path)
    if proc.returncode != 0:  # pragma: no cover
        raise RuntimeError("Bad return code from git... {} add better exception handling here.."
                           .format(proc.returncode))
    return proc.stdout.splitlines()


def git_status_ui(path, *args):  # pragma: no cover
    # For unittesting purposes, this function is a nuisance
    if unitesting:
        return
    # Don't redirect the std* streams; let the output go straight to the console
    cmd = [GIT_BIN, "status", "."]
    cmd.extend(args)
    call(cmd, cwd=path)



from __future__ import unicode_literals

from ksconf.consts import KSCONF_DEBUG


def _xargs(iterable, cmd_len=1024):
    fn_len = 0
    buf = []
    iterable = list(iterable)
    while iterable:
        s = iterable.pop(0)
        l = len(s) + 1
        if fn_len + l >= cmd_len:
            yield buf
            buf = []
            fn_len = 0
        buf.append(s)
        fn_len += l
    if buf:
        yield buf


def debug_traceback():  # pragma: no cover
    """ If the 'KSCONF_DEBUG' environmental variable is set, then show a stack trace. """
    level = 10
    from os import environ
    if KSCONF_DEBUG in environ:
        # TODO:  Pop one off the top of the stack to hide THIS function
        import traceback
        traceback.print_exc(level)



from __future__ import unicode_literals

# ANSI_COLOR = "\x1b[{0}m"
ANSI_BOLD = 1
ANSI_RED = 31
ANSI_GREEN = 32
ANSI_YELLOW = 33
ANSI_RESET = 0
TTY_COLOR_MODE = "auto"


class TermColor:
    """
    Simple color setting helper class that's a context manager wrapper around a stream.
    This ensure that the color is always reset at the end of a session.
    """

    def __init__(self, stream):
        self.stream = stream

        mode = TTY_COLOR_MODE.lower()
        if mode in ("off", "no", "disable"):
            self.color_enabled = False
        elif mode in ("on", "force") or (hasattr(stream, "isatty") and stream.isatty()):
            self.color_enabled = True
        else:
            self.color_enabled = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def write(self, content):
        return self.write(content)

    def color(self, *codes):
        if codes and self.color_enabled:
            self.stream.write("\x1b[{}m".format(";".join([str(i) for i in codes])))

    def reset(self):
        self.color(ANSI_RESET)



"""
Silly simple Python version compatibility items
"""

import sys

# Since 'list' in Python 3.7 doesn't support __class_getitem__ (See PEP 560)
# We could just always import from typing, but that appears to cause some issues
# with type hinting, and this appears to work correctly.
if sys.version_info < (3, 8):
    from typing import List, Tuple
else:
    List = list
    Tuple = tuple


try:
    # Python 3.9 and later
    from functools import cache
except ImportError:
    from functools import lru_cache
    cache = lru_cache(maxsize=None)


def handle_py3_kw_only_args(kw, *default_args):
    """ Fake support for Python 3.8+ style keyword-only style arguments, or ``*`` arg syntax.

    Example Python 3.8+ syntax:

    ..  code-block:: py

        def f(arg, *args, a=True, b=False):
            ...

    Example Python 3.7 (and earlier) syntax with this helper function:

    ..  code-block:: py

        def f(arg, *args, **kw_only):
            a, b = handle_py3_kw_only_args(kw_only, ("a", True), ("b", False))
            ...

    :param dict kw: keyword arguments provided to the calling function. Be aware
                    that this dict will be empty after this function is done.
    :param tuple default_args: pairs of keyword argument to the caller function
                               in argument (arg_name, default_value) order.
    :raises TypeError: if ``kw`` contains any keys not defined in ``args``
                       This mirrors Python's native behavior when an unexpected
                       argument is passed to a function.
    """
    out = []
    for arg_name, arg_default in default_args:
        try:
            out.append(kw.pop(arg_name))
        except KeyError:
            out.append(arg_default)
    if kw:
        import inspect
        caller = inspect.currentframe().f_back.f_code.co_name
        # Should all unexpected args be reported?  feels like this good enough
        raise TypeError("{} got an unexpected keyword argument '{}'"
                        .format(caller, list(kw)[0]))
    return out


del sys

__all__ = [
    "List",
    "Tuple",
    "cache",
    "handle_py3_kw_only_args",
]



from __future__ import annotations

import json
from datetime import datetime, timedelta
from pathlib import Path, PurePath
from shutil import copy2, rmtree

from ksconf.builder import BuildCacheException
from ksconf.util.file import file_hash


class FileSet:
    """ A collection of fingerprinted files.

    Currently the fingerprint is only a SHA256 hash.

    Two constructors are provided for building an instance from either file that
    live on the filesystem, via :py:meth:`from_filesystem` or from a persisted
    cached record available from the :py:meth:`from_cache`.
    The filesystem version actively reads all inputs files at object creation
    time, so this can be costly, especially if repeated.
    """
    # XXX: Do we need both files (set), and file_meta (dict)?  Try to make this work with just files_meta
    __slots__ = ["files", "files_meta"]

    def __init__(self):
        self.files = set()
        self.files_meta = {}

    def __eq__(self, other: "FileSet") -> bool:
        return self.files_meta == other.files_meta

    def __ne__(self, other: "FileSet") -> bool:
        return self.files_meta != other.files_meta

    '''
    def __iadd__(self, other):
        self.files.update(other.files)
        self.files_meta.update(other.files_meta)

    def __add__(self, other):
        combined = FileSet()
        for item in (self, other):
            combined.files.update(item.files)
            combined.files_meta.update(item.files_meta)
        return combined
    '''

    def __iter__(self):
        return iter(self.files)

    def __len__(self):
        return len(self.files)

    @classmethod
    def from_filesystem(cls, root, files):
        instance = cls()
        root = Path(root)
        for file_name in files:
            # XXX: Support globs
            if file_name.endswith("/"):
                # Recursive directory walk
                instance.add_glob(root, file_name + "**/*")
            elif "*" in file_name:
                instance.add_glob(root, file_name)
            else:
                instance.add_file(root, file_name)
        return instance

    @classmethod
    def from_cache(cls, data):
        instance = cls()
        for file_name, meta in data.items():
            file_name = PurePath(file_name)
            instance.files.add(file_name)
            instance.files_meta[file_name] = meta
        return instance

    def add_file(self, root, relative_path):
        """ Add a simple relative path to a file to the FileSet. """
        relative_path = PurePath(relative_path)
        p = root / relative_path
        if not p.is_file():
            if p.is_dir():
                # Audience: Exception text relevant to from_filesystem() caller
                raise BuildCacheException(
                    f"Expected file '{p}' is actually a directory.  If this is "
                    "correct indicate a directory with a trailing slash: "
                    f"'{p}/'")
            raise BuildCacheException(f"Missing expected file {p}")
        fp = self.get_fingerprint(p)
        self.files.add(relative_path)
        self.files_meta[relative_path] = fp

    def add_glob(self, root, pattern):
        """ Recursively add all files matching glob pattern. """
        for p in root.glob(pattern):
            if p.is_file():
                relative_path = p.relative_to(root)

                fp = self.get_fingerprint(p)
                self.files.add(relative_path)
                self.files_meta[relative_path] = fp

    @staticmethod
    def get_fingerprint(path):
        return {
            "hash": file_hash(path)
        }

    def copy_all(self, src_dir, dest_dir):
        """ Copy a the given set of files from one location to another. """
        src_dir = Path(src_dir)
        dest_dir = Path(dest_dir)
        for file_name in self.files:
            src = src_dir / file_name
            dest = dest_dir / file_name
            if not dest.parent.is_dir():
                dest.parent.mkdir(parents=True)
            copy2(src, dest)


class CachedRun:
    __slots__ = ["root", "config_file", "cache_dir", "_info", "_settings", "_state"]

    STATE_NEW = "new"
    STATE_EXISTS = "exists"
    STATE_TAINT = "taint"
    STATE_DISABLED = "disabled"

    _timestamp_format = "%Y-%m-%d %H:%M:%S"

    def __init__(self, root: Path):
        self.root = root
        self.config_file = self.root / "cache.json"
        self.cache_dir = self.root / "data"
        self._settings = {}
        self._info = {}
        self._state = self.STATE_NEW

        if not self.cache_dir.is_dir():
            self.cache_dir.mkdir(parents=True)
        elif self.config_file.is_file() and self.cache_dir.is_dir():
            self._state = self.STATE_EXISTS

    def set_settings(self, cache_settings):
        self._settings.update(cache_settings)

    @property
    def cached_inputs(self):
        return FileSet.from_cache(self._info["inputs"])

    @property
    def cached_outputs(self):
        return FileSet.from_cache(self._info["outputs"])

    @property
    def exists(self):
        return self._state in (self.STATE_EXISTS,)

    @property
    def is_new(self):
        return self._state == self.STATE_NEW

    @property
    def is_expired(self):
        timeout = self._settings["timeout"]
        if timeout is None:
            return False
        try:
            create_time = self._info["timestamp"]
            timeout = timedelta(seconds=timeout)
            expired = create_time + timeout
            return datetime.now() > expired
        except KeyError:  # no cover
            raise   # XXX: For initial testing
            # If anything about the info/settings were missing; assume expired
            return True

    @property
    def is_disabled(self):
        return self._state == self.STATE_DISABLED

    def inputs_identical(self, inputs: FileSet) -> bool:
        return self.cached_inputs == inputs

    def dump(self):
        def map_keys(d):
            return {str(k): v for k, v in d.items()}
        mode = "w"
        meta = dict(self._info)
        inputs = meta.pop("inputs")
        outputs = meta.pop("outputs")
        data = {
            "settings": self._settings,
            "timestamp": datetime.now().strftime(self._timestamp_format),
            "meta": meta,
            "state": {
                "inputs": map_keys(inputs),
                "outputs": map_keys(outputs),
            },
        }
        with self.config_file.open(mode) as f:
            json.dump(data, f, indent=2)

    def load(self):
        def map_keys(d):
            return {Path(k): v for k, v in d.items()}
        with self.config_file.open() as f:
            data = json.load(f)
        info = {}
        info.update(data["meta"])
        info["timestamp"] = datetime.strptime(data["timestamp"],
                                              self._timestamp_format)
        for state_type, value in data["state"].items():
            info[state_type] = map_keys(value)
        self._settings = data["settings"]
        self._info = info

    def set_cache_info(self, type: str, data: FileSet):
        assert type in ("inputs", "outputs")
        self._info[type] = data.files_meta

    def rename(self, dest):
        if dest.is_dir():
            rmtree(dest)
        self.root.rename(dest)
        # Update root incase any other operations need to take place
        self.root = dest

    def taint(self):
        cf = self.config_file
        if cf.exists():
            cf.unlink()
        self._state = self.STATE_TAINT

    def disable(self):
        self._state = self.STATE_DISABLED



# -*- coding: utf-8 -*-
"""
XXX: Split out conf, spec, binary file handlers into separate registerable (and therefore easily extendable) functions (using decorators, possibly)
XXX: Move the overwrite-as-necessary logic into a subclass; for several use cases we just don't care because 'target' is a brand new directory
"""

from __future__ import absolute_import, unicode_literals

import os
import re
import sys
from io import open

from ksconf.commands import ConfFileProxy
from ksconf.conf.delta import show_text_diff
from ksconf.conf.merge import merge_conf_files
from ksconf.conf.parser import PARSECONF_MID, PARSECONF_STRICT
from ksconf.consts import SMART_CREATE, SMART_NOCHANGE, SMART_UPDATE
from ksconf.layer import DirectLayerRoot, DotDLayerRoot, LayerConfig, LayerFilter, LayerRootBase
from ksconf.util.compare import file_compare
from ksconf.util.file import _is_binary_file, smart_copy


class LayerCombinerException(Exception):
    pass


class LayerCombiner:
    """
    Class to recursively combine layers (directories) into a single rendered output target directory.
    This is heavily used by the ``ksconf combine`` command as well as by the ``package`` command.


    Typical class use case:

    ::
        lc = LayerCombiner()

        # Setup source, either
            (1) lc.set_source_dirs()  OR
            (2) lc.set_layer_root()

    Call hierarch:

    ::

        lc.combine()                    Entry point
            -> prepare()                Directory, layer prep
                -> prepare_target_dir() Make dir; subclass handles marker here (combine CLI)
            -> pre_combine_inventory()  Hook for pre-processing (or alerting) the set of files to combine
            -> combine_files()          Main worker function
            -> post_combine()           Optional, cleanup leftover files
    """

    # Note this is case sensitive.  Don't be lazy, name your files correctly  :-)
    conf_file_re = re.compile(r"([a-z_-]+\.conf|(default|local)\.meta)$")
    spec_file_re = re.compile(r"\.conf\.spec$")

    def __init__(self,
                 follow_symlink: bool = False,
                 banner: str = "",
                 dry_run: bool = False,
                 quiet: bool = False):
        self.layer_root: LayerRootBase = None
        self.config = LayerConfig()
        self.layer_filter = LayerFilter()
        self.banner = banner
        self.dry_run = dry_run
        self.quiet = quiet

        self.config.follow_symlink = follow_symlink

        # Internal tracking variables
        self.layer_names_all = set()
        self.layer_names_used = set()

        # Not a great long-term design, but good enough for initial conversion from command-based design
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    def set_source_dirs(self, sources: list):
        self.layer_root = DirectLayerRoot(config=self.config)
        for src in sources:
            self.layer_root.add_layer(src)

    def set_layer_root(self, root):
        layer_root = DotDLayerRoot(config=self.config)
        layer_root.set_root(root)
        self.layer_root = layer_root

    def add_layer_filter(self, action, pattern):
        self.layer_filter.add_rule(action, pattern)

    def combine(self, target):
        """
        Combine layers into ``target`` directory.
        """
        layer_root = self.layer_root
        if layer_root is None:
            raise TypeError("Call either set_source_dirs() or set_layer_root() before calling combine()")
        self.prepare(target)
        # Build a common tree of all src files.
        src_file_listing = layer_root.list_files()
        src_file_listing = self.pre_combine_inventory(target, src_file_listing)
        self.combine_files(target, src_file_listing)
        self.post_combine(target)

    def prepare(self, target):
        """ Start the combine process.  This includes directory checking,
        applying layer filtering, and marker file handling. """
        layer_root, layer_filter = self.layer_root, self.layer_filter

        self.prepare_target_dir(target)

        self.layer_names_all.update(layer_root.list_layer_names())
        if layer_root.apply_filter(layer_filter):
            self.layer_names_used.update(layer_root.list_layer_names())
        else:
            self.layer_names_used.update(self.layer_names_all)

    def prepare_target_dir(self, target):
        """ Hook to ensure destination directory is ready for use.  This can be overridden
        to adder marker file handling for use cases that need it (e.g., the 'combine' command)
        """
        if not os.path.isdir(target):
            os.mkdir(target)

    def pre_combine_inventory(self, target, src_files):
        """ Hook point for pre-processing before any files are copied/merged """
        del target
        return src_files

    def post_combine(self, target):
        """ Hook point for post-processing after all copy/merge operations have been completed. """
        del target

    def combine_files(self, target, src_files):
        layer_root = self.layer_root

        def physical_paths(l: LayerRootBase.File):
            return [s.physical_path for s in l]

        for src_file in sorted(src_files):
            # Source file must be in sort order (10-x is lower prio and therefore replaced by 90-z)
            sources = list(layer_root.get_file(src_file))
            try:
                dest_fn = sources[0].logical_path
            except IndexError:
                raise LayerCombinerException("File disappeared during execution?  {src_file}\n")

            dest_path = os.path.join(target, dest_fn)

            # Make missing destination folder, if missing
            dest_dir = os.path.dirname(dest_path)
            if not os.path.isdir(dest_dir) and not self.dry_run:
                os.makedirs(dest_dir)

            # Determine handling method based on source count and filename pattern
            if len(sources) == 1:
                # Copy only file (most common case)
                method = "copy"
            elif self.spec_file_re.search(dest_fn):
                method = "concatenate"
            elif self.conf_file_re.search(dest_fn):
                method = "merge"
            else:
                # Copy highest precedence
                method = "copy"

            if method == "copy":
                # self.stderr.write(f"Considering {dest_fn:50}  NON-CONF Copy from source:  "
                #                   f"{sources[-1].physical_path!r}\n")
                # Always use the last file in the list (since last directory always wins)
                src_file = sources[-1].physical_path
                if self.dry_run:
                    if os.path.isfile(dest_path):
                        if file_compare(src_file, dest_path):
                            smart_rc = SMART_NOCHANGE
                        else:
                            if (_is_binary_file(src_file) or _is_binary_file(dest_path)):
                                # Binary files.  Can't compare...
                                smart_rc = "DRY-RUN (NO-DIFF=BIN)"
                            else:
                                show_text_diff(self.stdout, dest_path, src_file)
                                smart_rc = "DRY-RUN (DIFF)"
                    else:
                        smart_rc = "DRY-RUN (NEW)"
                else:
                    smart_rc = smart_copy(src_file, dest_path)
                if smart_rc != SMART_NOCHANGE:
                    if not self.quiet:
                        self.stderr.write(f"Copy <{smart_rc}>   {dest_path:50}  from {src_file}\n")
                del src_file

            elif method == "merge":
                try:
                    # Handle merging conf files
                    dest = ConfFileProxy(dest_path, "r+",
                                         parse_profile=PARSECONF_MID)
                    srcs = [ConfFileProxy(s.physical_path, "r",
                                          parse_profile=PARSECONF_STRICT) for s in sources]
                    # self.stderr.write(f"Considering {dest_fn:50}  CONF MERGE from source:  "
                    #                   f"{1!sources[0].physical_path}\n")
                    smart_rc = merge_conf_files(dest, srcs, dry_run=self.dry_run,
                                                banner_comment=self.banner)
                    if smart_rc != SMART_NOCHANGE:
                        if not self.quiet:
                            self.stderr.write(f"Merge <{smart_rc}>   {dest_path:50}  "
                                              f"from {physical_paths(sources)!r}\n")
                finally:
                    # Protect against any dangling open files:  (ResourceWarning: unclosed file)
                    dest.close()
                    for src in srcs:
                        src.close()
                    del srcs, dest

            elif method == "concatenate":
                combined_content = ""
                last_mtime = max(src.mtime for src in sources)
                for src in sources:
                    with open(src, "r") as stream:
                        content = stream.read()
                        if not content.endswith("\n"):
                            content += "\n"
                        combined_content += content
                        del content
                smart_rc = SMART_CREATE
                if os.path.isfile(dest_path):
                    with open(dest_path) as stream:
                        dest_content = stream.read()
                    if dest_content == combined_content:
                        smart_rc = SMART_NOCHANGE
                    else:
                        smart_rc = SMART_UPDATE
                    del dest_content

                if not self.dry_run:
                    with open(dest_path, "w") as stream:
                        stream.write(combined_content)

                if smart_rc != SMART_NOCHANGE:
                    if not self.quiet:
                        self.stderr.write(f"Concatenate <{smart_rc}>   {dest_path:50}  "
                                          f"from {physical_paths(sources)!r}\n")
                os.utime(dest_path, (last_mtime, last_mtime))
                del combined_content
            else:
                raise AssertionError(f"Internal implementation error.  Unknown method={method}")



""" Defines all command prompt entry points for CLI actions

This is a silly hack allows for fallback mechanism when
    (a) running unit tests (can happen before install)
    (b) if entrypoints or pkg_resources are not available at run time (Splunk's embedded python)
"""

from __future__ import absolute_import, unicode_literals

from importlib import import_module
from typing import NamedTuple


class Ep(NamedTuple):
    name: str
    module_name: str
    object_name: str


# autopep8: off
_entry_points = {
    "console_scripts": [
        Ep("ksconf", "ksconf.__main__", "cli"),
    ],
    # Custom end_point for ksconf subcommand registration
    "ksconf_cmd": [
        Ep("check",         "ksconf.commands.check",        "CheckCmd"),
        Ep("combine",       "ksconf.commands.combine",      "CombineCmd"),
        Ep("diff",          "ksconf.commands.diff",         "DiffCmd"),
        Ep("package",       "ksconf.commands.package",      "PackageCmd"),
        Ep("filter",        "ksconf.commands.filter",       "FilterCmd"),
        Ep("promote",       "ksconf.commands.promote",      "PromoteCmd"),
        Ep("merge",         "ksconf.commands.merge",        "MergeCmd"),
        Ep("minimize",      "ksconf.commands.minimize",     "MinimizeCmd"),
        Ep("snapshot",      "ksconf.commands.snapshot",     "SnapshotCmd"),
        Ep("sort",          "ksconf.commands.sort",         "SortCmd"),
        Ep("rest-export",   "ksconf.commands.restexport",   "RestExportCmd"),
        Ep("rest-publish",  "ksconf.commands.restpublish",  "RestPublishCmd"),
        Ep("unarchive",     "ksconf.commands.unarchive",    "UnarchiveCmd"),
        Ep("xml-format",    "ksconf.commands.xmlformat",    "XmlFormatCmd"),
    ],
}
# autopep8: on


def get_entrypoints_setup():
    setup = {}
    for (group, entries) in _entry_points.items():
        setup[group] = [f"{ep.name} = {ep.module_name}:{ep.object_name}" for ep in entries]
    return setup


class LocalEntryPoint:
    """ Bare minimum stand-in for entrypoints.EntryPoint """

    def __init__(self, data):
        self._data = data
        self.dist = None

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def load(self):
        mod = import_module(self.module_name)
        return getattr(mod, self.object_name)


def get_entrypoints_fallback(group):
    entry_points = {}
    for ep in _entry_points[group]:
        entry_points[ep.name] = LocalEntryPoint(ep)
    return entry_points


def debug():
    # For debugging internally defined entrypoints
    print("Builtin entrypoints:")
    for (_, entries) in _entry_points.items():
        print("[group]")
        for ep in entries:
            print(f"{ep.name:15} = {ep.module_name:30} : {ep.object_name}")
        print("")


if __name__ == '__main__':
    debug()



"""
SUBCOMMAND:  ``ksconf snapshot --output=FILE.json <PATH> [ ... <PATH-n> ]``

Usage example:

.. code-block:: sh

    ksconf snapshot --output=daily.json /opt/splunk/etc/app/

"""
from __future__ import absolute_import, unicode_literals

import json
import os
import sys
from argparse import FileType

from ksconf import __vcs_info__, __version__
from ksconf.commands import KsconfCmd, dedent
from ksconf.conf.parser import GLOBAL_STANZA, PARSECONF_MID_NC, ConfParserException, parse_conf
from ksconf.consts import EXIT_CODE_NO_SUCH_FILE, EXIT_CODE_SUCCESS
from ksconf.util.completers import DirectoriesCompleter, FilesCompleter
from ksconf.util.file import file_hash


class ConfSnapshotConfig:
    max_file_size = 10 * 1024 * 1024

    # max_files = 10000
    # include_parts = [ "conf", "meta", "lookups", "data/ui", "data/model", "data/ui/nav" ]


class ConfSnapshot:
    schema_version = 0.2

    def __init__(self, config):
        self.config = config
        self._data = []

    @staticmethod
    def _decode_path_meta(path):
        meta = {}
        if path.endswith(".conf"):
            meta["type"] = "conf"
            meta["conf"] = os.path.basename(path)[:-5]
        elif path.endswith(".meta"):
            meta["type"] = "metadata"
        # meta["level"] = local vs default
        # meta["folder"] = apps, deployment-apps, master-apps, users, ...
        # Eventually need some kind of "data/*" decoder as well.
        # XXX: Extract the app name ...
        return meta

    @staticmethod
    def _get_file_info(path):
        file_info = {}
        file_info["path"] = path
        # XXX:  Format as ISO something or other (with timezone, ugh); for now just use int
        st = os.stat(path)
        file_info["mtime"] = int(st.st_mtime)
        file_info["size"] = st.st_size
        # Full hash is probably an overkill; first 20 characters is probably plenty; and this is a
        # hash of the RAW file, not the config contents.
        file_info["hash"] = file_hash(path)
        # For now, don't bother capturing OS specific stuff like mode or file ownership.
        # Let's assume that if we can read it, Splunk can read it too.
        return file_info

    def snapshot_file_conf(self, path):
        # XXX: If we are unable to read the file (IOError/OSError) that should be reported via
        # metadata or via 'failure'
        record = {}
        record["meta"] = self._decode_path_meta(path)
        record["file"] = self._get_file_info(path)
        try:
            data = parse_conf(path, profile=PARSECONF_MID_NC)
            # May need to format this differently.   Specifically need some way to textually
            # indicate the global stanza
            conf = record["conf"] = []
            for (stanza, stanza_data) in data.items():
                rec = {
                    "stanza": stanza,
                    "attributes": stanza_data
                }
                if stanza is GLOBAL_STANZA:
                    rec["stanza"] = "**GLOBAL_STANZA**"
                conf.append(rec)
        except ConfParserException as e:
            record["conf"] = None
            record["failure"] = f"{e}"
        self._data.append(record)

    def snapshot_dir(self, path):
        path = os.path.abspath(path)
        for (dirpath, dirnames, filenames) in os.walk(path):
            # Remove unwanted directories
            for avoid in ("bin", "static"):
                if avoid in dirnames:
                    dirnames.remove(avoid)
            for fn in filenames:
                if fn.endswith(".conf") or fn.endswith(".meta"):
                    self.snapshot_file_conf(os.path.join(path, dirpath, fn))

    def write_snapshot(self, stream, **kwargs):
        record = {
            "schema_version": self.schema_version,
            "software": {
                "name": "ksconf",
                "version": [__version__, __vcs_info__],
                "command": sys.argv,
            },
        }
        record["records"] = self._data
        json.dump(record, stream, **kwargs)


class SnapshotCmd(KsconfCmd):
    help = "Snapshot .conf file directories into a JSON dump format"
    description = dedent("""\
    Build a static snapshot of various configuration files stored within a structured json export
    format.  If the .conf files being captured are within a standard Splunk directory structure,
    then certain metadata and namespace information is assumed based on typical path locations.
    Individual apps or conf files can be collected as well, but less metadata may be extracted.
    """)

    def register_args(self, parser):
        parser.add_argument("path", metavar="PATH", nargs="+", type=str,
                            help="Directory from which to load configuration files.  "
                                 "All .conf and .meta file are included recursively."
                            ).completer = DirectoriesCompleter()
        parser.add_argument("--output", "-o", metavar="FILE",
                            type=FileType("w"), default=self.stdout,
                            help=dedent("""\
            Save the snapshot to the named files.  If not provided, the snapshot is written to
            standard output.""")
                            ).completer = FilesCompleter(allowednames=["*.json"])
        parser.add_argument("--minimize", action="store_true", default=False,
                            help="Reduce the size of the JSON output by removing whitespace.  "
                            "Reduces readability.")

    def run(self, args):
        ''' Snapshot multiple configuration files into a single json snapshot. '''
        cfg = ConfSnapshotConfig()
        confSnap = ConfSnapshot(cfg)
        for path in args.path:
            if os.path.isfile(path):
                confSnap.snapshot_file_conf(path)
            elif os.path.isdir(path):
                confSnap.snapshot_dir(path)
            else:
                self.stderr.write(f"No such file or directory {path}\n")
                return EXIT_CODE_NO_SUCH_FILE

        if args.minimize:
            confSnap.write_snapshot(args.output)
        else:
            confSnap.write_snapshot(args.output, indent=2)
        return EXIT_CODE_SUCCESS



# -*- coding: utf-8 -*-
""" Splunk App content inventory and signature management

"""

from __future__ import absolute_import, annotations, unicode_literals

import hashlib
import json
from dataclasses import asdict, dataclass, field
from os import fspath
from pathlib import Path, PurePosixPath
from typing import Iterable

from ksconf.archive import extract_archive
from ksconf.compat import List
from ksconf.consts import MANIFEST_HASH, UNSET
from ksconf.util.file import file_hash, relwalk


class AppArchiveError(Exception):
    pass


class AppArchiveContentError(Exception):
    """ Problem with the contents of an archive """
    pass


class AppManifestStorageError(Exception):
    pass


class AppManifestStorageInvalid(AppManifestStorageError):
    pass


@dataclass(order=True)
class AppManifestFile:
    path: PurePosixPath
    mode: int
    size: int
    hash: str = None

    def content_match(self, other):
        return self.hash == other.hash

    def to_dict(self):
        d = asdict(self)
        d["path"] = fspath(self.path)
        return d

    @classmethod
    def from_dict(cls, data: dict) -> "AppManifestFile":
        return cls(PurePosixPath(data["path"]), data["mode"], data["size"], data["hash"])


@dataclass
class AppManifest:
    name: str = None
    source: str = None
    hash_algorithm: str = field(default=MANIFEST_HASH)
    _hash: str = field(default=UNSET, init=False)
    files: List[AppManifestFile] = field(default_factory=list)

    def __eq__(self, other: AppManifest) -> bool:
        if self.name != other.name or self.hash != other.hash:
            return False
        return sorted(self.files) == sorted(other.files)

    @property
    def hash(self):
        # How do we trigger an update if files is updated?  Should we make the caller worry about this?  Maybe call a reset() method?
        # Should we convert the list to a tuple as soon as hash is calculated for the first time, then require cloning to make further modifications?
        if self._hash is UNSET:
            self._hash = self._calculate_hash()
        return self._hash

    def _calculate_hash(self) -> str:
        """ Build unique hash based on file content """
        # Path sort order notes.  Sorting based on Path objects is different
        # than textual sorting, consider:
        #   README.txt
        #   README/inputs.conf.spec
        # Sorting based on path is equivalent to sorting tuples of path components
        # Like doing sort(key=lambda s: s.path.split("/"))
        parts = []
        for f in sorted(self.files):
            # If one or more hash is None, then refuse to calculate hash
            if f.hash is None:
                return None
            parts.append(f"{f.hash} 0{f.mode:o} {'/'.join(f.path.parts)}")
        parts.insert(0, self.name)
        payload = "\n".join(parts)
        h = hashlib.new(self.hash_algorithm)
        h.update(payload.encode("utf-8"))
        return h.hexdigest()

    @classmethod
    def from_dict(cls, data: dict) -> "AppManifest":
        files = [AppManifestFile.from_dict(f) for f in data["files"]]
        o = cls(data["name"], source=data["source"],
                hash_algorithm=data["hash_algorithm"], files=files)
        o._hash = data["hash"]
        return o

    def to_dict(self):
        d = {
            "name": self.name,
            "source": self.source,
            "hash_algorithm": self.hash_algorithm,
            "hash": self.hash,
            "files": [f.to_dict() for f in self.files]
        }
        return d

    @classmethod
    def from_archive(cls, archive: Path,
                     calculate_hash=True) -> "AppManifest":
        """
        Create as new AppManifest from a tarball.  Set ``calculate_hash`` as
        False when only a file listing is needed.
        """
        manifest = cls(source=fspath(archive))
        app_names = set()
        archive = Path(archive)

        if calculate_hash:
            h_ = hashlib.new(cls.hash_algorithm)

            def gethash(content):
                h = h_.copy()
                h.update(content)
                return h.hexdigest()
        else:
            def gethash(_):
                return None

        for gaf in extract_archive(archive, lambda _: calculate_hash):
            app, relpath = gaf.path.split("/", 1)
            app_names.add(app)
            hash = gethash(gaf.payload)
            f = AppManifestFile(PurePosixPath(relpath), gaf.mode, gaf.size, hash)
            manifest.files.append(f)
        if len(app_names) > 1:
            raise AppArchiveContentError("Found multiple top-level app names!  "
                                         f"Archive {archive} contains apps {', '.join(app_names)}")
        manifest.name = app_names.pop()
        return manifest

    @classmethod
    def from_filesystem(cls, path: Path,
                        name: str = None,
                        follow_symlinks=False,
                        calculate_hash=True) -> "AppManifest":
        """
        Create as new AppManifest from an existing directory structure.
        Set ``calculate_hash`` as False when only a file listing is needed.
        """
        path = Path(path)
        if name is None:
            name = path.name
        manifest = cls(name, source=path)
        h_ = hashlib.new(cls.hash_algorithm)

        for (root, _, files) in relwalk(path, followlinks=follow_symlinks):
            root_path = PurePosixPath(root)
            for file_name in files:
                rel_path: PurePosixPath = root_path.joinpath(file_name)
                full_path: Path = path.joinpath(root, file_name)
                st = full_path.stat()
                amf = AppManifestFile(rel_path, st.st_mode & 0o777, st.st_size)
                if calculate_hash:
                    h = h_.copy()
                    h.update(full_path.read_bytes())
                    amf.hash = h.hexdigest()
                    del h
                manifest.files.append(amf)
        return manifest

    def find_local(self) -> Iterable[AppManifestFile]:
        for f in self.files:
            if f.path.parts[0] == "local" or f.path.name == "local.meta":
                yield f

    def check_paths(self):
        """ Check for dangerous paths in the archive. """
        for file in self.files:
            path = file.path
            if path.is_absolute():
                raise AppArchiveContentError(f"Found an absolute path {file.path}")
            if ".." in path.parts or path.parts[0].startswith("~"):
                raise AppArchiveContentError(f"Found questionable path manipulation in '{file.path}'")


@dataclass
class StoredArchiveManifest:
    """
    Stored manifest for a tarball.  Typically the manifest file lives in the
    same directory as the archive.  Details around the naming, storage, and
    clean up of these persistent manifest files are managed by the caller.
    """
    archive: Path
    size: int
    mtime: float
    hash: str
    _manifest_dict: dict = field(init=False)
    _manifest: AppManifest = field(init=False, default=None)

    @property
    def manifest(self) -> AppManifest:
        # Lazy load the full manifest details later (after size/mtime/hash have been confirmed)
        if self._manifest is None:
            if self._manifest_dict:
                try:
                    self._manifest = AppManifest.from_dict(self._manifest_dict)
                except KeyError as e:
                    raise AppManifestStorageError(f"Error loading manifest {e}")

        return self._manifest

    @classmethod
    def from_dict(cls, data: dict) -> "StoredArchiveManifest":
        o = cls(Path(data["archive"]), data["size"], data["mtime"], data["hash"])
        o._manifest_dict = data["manifest"]
        return o

    def to_dict(self):
        return {
            "archive": fspath(self.archive),
            "size": self.size,
            "mtime": self.mtime,
            "hash": self.hash,
            "manifest": self.manifest.to_dict(),
        }

    def write_json_manifest(self, manifest_file: Path):
        data = self.to_dict()
        with open(manifest_file, "w") as fp:
            json.dump(data, fp, indent=1)

    @classmethod
    def read_json_manifest(cls, manifest_file: Path) -> "StoredArchiveManifest":
        with open(manifest_file) as fp:
            data = json.load(fp)
        return cls.from_dict(data)

    @classmethod
    def from_file(cls,
                  archive: Path,
                  manifest: AppManifest
                  ) -> "StoredArchiveManifest":
        """
        Construct instance from a tarball.
        """
        stat = archive.stat()

        hash = file_hash(archive, MANIFEST_HASH)
        o = cls(archive, stat.st_size, stat.st_mtime, hash)
        o._manifest = manifest
        return o

    @classmethod
    def from_json_manifest(cls,
                           archive: Path,
                           stored_file: Path) -> "StoredArchiveManifest":
        """
        Attempt to load as stored manifest from archive & stored manifest paths.
        If the archive has changed since the manifest was stored, then an
        exception will be raised indicating the reason for invalidation.
        """
        # XXX: Optimization: tests if archive is newer than stored_file.  No need to open/parse JSON.
        if not stored_file:
            raise AppManifestStorageInvalid("No stored manifest found")

        try:
            stored = cls.read_json_manifest(stored_file)

            if stored.archive != archive:
                raise AppManifestStorageInvalid(f"Archive name differs: {stored.archive!r} != {archive!r}")
            stat = archive.stat()
            if stored.size != stat.st_size:
                raise AppManifestStorageInvalid(f"Archive file size differs:  {stored.size} != {stat.st_size}")
            if abs(stored.mtime - stat.st_mtime) > 0.1:
                raise AppManifestStorageInvalid(f"Archive file mtime differs: {stored.mtime} vs {stat.st_mtime}")
        except (ValueError, KeyError) as e:
            raise AppManifestStorageError(f"Unable to load stored manifest due to {e}")
        return stored


def get_stored_manifest_name(archive: Path) -> Path:
    """
    Calculate the name of the stored manifest file based on ``archive``.
    """
    c = archive.with_name(f".{archive.name}.manifest")
    return c


def create_manifest_from_archive(
        archive_file: Path,
        manifest_file: Path,
        manifest: AppManifest) -> StoredArchiveManifest:
    """
    Create a new stored manifest file based on a given archive.
    """
    if manifest_file is None:
        manifest_file = get_stored_manifest_name(archive_file)
    sam = StoredArchiveManifest.from_file(archive_file, manifest)
    sam.write_json_manifest(manifest_file)
    return sam


def load_manifest_for_archive(
        archive: Path,
        manifest_file: Path = None,
        read_manifest=True,
        write_manifest=True) -> AppManifest:
    """
    Load manifest for ``archive`` and create a stored copy of the manifest in
    ``manifest_file``.  On subsequent calls the manifest data stored to disk
    will be reused assuming ``manifest_file`` is up-to-date.

    File modification time and size are used to determine if ``archive`` has
    been changed since the ``manifest_file`` was written.

    If no ``manifest_file`` is provided, the default manifest naming convention
    will be applied where the ``manifest_file`` is stored in the same directory
    as ``archive``.
    """
    # XXX: Add optimization to check if archive is newer than manifest_file, assume old
    archive = Path(archive)
    manifest = None

    if manifest_file is None and read_manifest or write_manifest:
        manifest_file = get_stored_manifest_name(archive)

    if read_manifest and manifest_file.exists():
        try:
            sam = StoredArchiveManifest.from_json_manifest(archive, manifest_file)
            manifest = sam.manifest
        except AppManifestStorageError as e:
            print(f"WARN:   loading stored manifest failed:  {e}")

    if manifest is None:
        print(f"Calculating manifest for {archive}")
        manifest = AppManifest.from_archive(archive)

        # We assume that a previously stored manifest has already undergone the
        # path check.  Checks should be redone from within the extraction process.
        manifest.check_paths()

        if write_manifest:
            create_manifest_from_archive(archive, manifest_file, manifest)

    return manifest



# Version file autogenerated by the build process
version = '0.11.1'
build = 1132
vcs_info = 'Git SHA1 bddd6219 committed on 2023-05-13'
package_name = 'ksconf'

if __name__ == '__main__':
    print('KSCONF_VERSION="0.11.1"\nKSCONF_BUILD="1132"\nKSCONF_VCS_INFO="Git SHA1 bddd6219 committed on 2023-05-13"')



# -*- coding: utf-8 -*-
"""


Cache build requirements:

    * Caching mechanism should inspect 'inputs' (collect file hashes) to determine if any content has
      changed.  If input varies, then command should be re-run.
    * Command (decorated function) should be generally unaware of all other details of build process,
      and it should *ONLY* be able to see files listed in "inputs"
    * Allow caching to be fully disabled (run in-place with no dir proxy) for CI/CD
    * Cache should have allow a timeout parameter


decorator used to implement caching:
    * decorator args:
        * inputs:       list or glob
        * outputs       (do we need this, can we just detect this??)
                        Default to "." (everything)
        * timeout=0     Seconds before cache should be considered stale
        * name=None     If not given, default to the short name of the function.
                        (Cache "slot"), must be filesystem safe]
"""

from functools import wraps
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Union

from ksconf.builder import QUIET, VERBOSE, BuildCacheException, BuildStep
from ksconf.builder.cache import CachedRun, FileSet


def _get_function_sourcecode_hash(f):
    import hashlib
    import inspect
    h = hashlib.new("sha256")
    code = inspect.getsource(f).encode("utf-8")
    h.update(code)
    return h.hexdigest()


class BuildManager:
    """ Management of individual build steps

    .. versionadded:: v0.8.0
    """

    def __init__(self):
        self.source_path = None
        self.build_path = None
        self.dist_path = None
        self.cache_path = None
        self._cache_enabled = True
        self._taint = False
        self._folder_set = False

    def taint_cache(self):
        self._taint = True

    def disable_cache(self):
        self._cache_enabled = False

    def get_build_step(self, output=None):
        kw = {}
        if output:
            kw["output"] = output
        step = BuildStep(self.build_path, self.source_path, self.dist_path, **kw)
        return step

    def set_folders(self, source_path, build_path, dist_path=None):
        self._folder_set = True
        self.source_path = Path(source_path).absolute()
        self.build_path = Path(build_path).absolute()
        if dist_path:
            self.dist_path = Path(dist_path).absolute()
        else:
            # Assume 'dist' lives along side 'build'
            self.dist_path = self.build_path.with_name("dist")
        self.cache_path = self.build_path.with_suffix(".cache")

    def is_folders_set(self):
        return self._folder_set

    def get_cache_info(self, name):
        path = self.cache_path / name
        cache_info = CachedRun(path)
        if self._taint:
            cache_info.taint()
        if not self._cache_enabled:
            cache_info.disable()
        return cache_info

    def cache(self, inputs: List[str], outputs: int,
              timeout: int = None,
              name: str = None,
              cache_invalidation: Union[dict, list, str] = None) -> None:
        """ function decorator for caching build steps
        Wrapped function must accept BuildStep instance as first parameters

        XXX:  Clearly document what things are good cache candidates and which are not.

        Example:

            * No extra argument to the function (at least currently)
            * Changes to inputs files are not supported
            * Deleting files aren't supported
            * Can only operate in a single directory given a limited set of inputs
            * Cannot read from the source directory, and agrees not to write to dist
              (In other words, limit all activities to build_path for deterministic behavior)
        """
        name_ = name
        cache_settings = {
            "name": name,
            "inputs": inputs,
            "outputs": outputs,
            "timeout": timeout,
            "cache_invalidation": cache_invalidation,
        }

        def decorator(f):
            # nonlocal name
            if name_ is None:
                name = f.__name__.split(".")[-1]

            f_source_hash = _get_function_sourcecode_hash(f)
            cache_settings["function_code_hash"] = f_source_hash

            @wraps(f)
            def wrapper(build_step: BuildStep) -> None:
                log = build_step.get_logger(name)
                use_cache = True
                cache = self.get_cache_info(name)

                # After the first cached build step is executed, block access to source path
                build_step.source_path = None

                if cache.is_disabled:
                    log("Caching disabled", VERBOSE)
                    # Run directly with no additional
                    return f(build_step)

                if cache.exists:
                    try:
                        cache.load()
                    except Exception as e:
                        log(f"Failed to load cache:  {e}", QUIET)
                        use_cache = False
                if cache._settings:
                    # Always use the most up-to-date value for timeout
                    cache.set_settings({"timeout": timeout})
                    try:
                        for setting in cache_settings:
                            if cache_settings[setting] != cache._settings[setting]:
                                log(f"Cache invalided due to setting '{setting}': "
                                    f"{cache_settings[setting]} was "
                                    f"{cache._settings[setting]}")
                                use_cache = False
                                break
                    except KeyError as e:
                        log(f"Cache invalided due to missing setting {e}")
                        use_cache = False
                # TODO: Check for cache tampering (confirm that existing files haven't
                #       been modified); user requestable
                current_inputs = FileSet.from_filesystem(self.source_path, inputs)
                # Determine if previous cache entry exists, and if the input files are the same
                if not cache.exists:
                    log("No cache found", VERBOSE)
                    use_cache = False
                elif not cache.inputs_identical(current_inputs):
                    # XXX: Tell user which file(s) changed?
                    log("Cache skipped due to change in inputs.  Will re-run.")
                    use_cache = False
                elif cache.is_expired:
                    log("Cache expired.  Will re-run.")
                    use_cache = False
                if use_cache:
                    # Cache HIT:  Reuse cache, by simply copying the outputs to the build folder
                    log("Cache used")
                    cached_output = cache.cached_outputs
                    cached_output.copy_all(cache.cache_dir, self.build_path)
                    log(f"Reused {len(cached_output)} output objects from cache", VERBOSE)
                    # XXX:  VERBOSE 3 should list all expanded files
                else:
                    # Cache MISS: Prepare to call the wrapped function
                    # 1. √ Make temporary location
                    # 2. √ Copy inputs to temp locations
                    # 3. √ Run wrapped functionality
                    # 4. √ Expand output into FileSet (Collect)
                    # 5. √ Store metadata (inputs/outputs/...) to JSON
                    # 6. √ Copy output to real build folder

                    # Make temporary folder for executing wrapped function
                    with TemporaryDirectory(dir=self.cache_path,
                                            prefix=f"{name}-tmp-") as temp_dir:
                        # NOTE: Make a 't' dir under the temp folder so that the cache.rename() doesn't remove the
                        #       folder managed by TemporaryDirectory, which it doesn't like.
                        temp_dir = Path(temp_dir) / "t"
                        final_cache_root = cache.root
                        cache = CachedRun(temp_dir)
                        alt_bs = build_step.alternate_path(cache.cache_dir)
                        # XXX: Copy any other settings from the original 'build_step' to our copied version
                        # Collect inputs from the source directory and copy them to the temporary directory
                        fs_inputs = FileSet.from_filesystem(self.source_path, inputs)
                        fs_inputs.copy_all(self.source_path, cache.cache_dir)
                        log(f"Copied {len(fs_inputs)} input files", VERBOSE * 2)
                        log(f"Copied input files: {', '.join(str(p) for p in fs_inputs)}", VERBOSE * 3)
                        try:
                            # Run wrapped function
                            ret = f(alt_bs)
                            if ret is not None:
                                raise NotImplementedError("A return value not supported for cached build steps")
                        except Exception as e:
                            log(f"Failed during executing.  Error: {e}", QUIET * 2)
                            # XXX: More error handling here.  Add custom exception class
                            raise
                        # TODO: Check for change to inputs.  This could lead to nondeterministic results
                        fs_inputs2 = FileSet.from_filesystem(cache.cache_dir, inputs)
                        if fs_inputs != fs_inputs2:
                            log("Inputs changed during execution", QUIET * 2)
                            raise BuildCacheException("Inputs were modified")
                        fs_outputs = FileSet.from_filesystem(cache.cache_dir, outputs)
                        # Store input/output fingerprint to the internal cache state and persist to disk
                        log(f"Capture {len(fs_outputs)} outputs")
                        cache.set_settings(cache_settings)
                        cache.set_cache_info("inputs", fs_inputs)
                        cache.set_cache_info("outputs", fs_outputs)
                        cache.dump()
                        # Copy output files to real build directory
                        fs_outputs.copy_all(cache.cache_dir, self.build_path)
                        cache.rename(final_cache_root)
                # No return (on purpose); as we don't want to track extra state

            wrapper.inputs = inputs
            wrapper.outputs = outputs
            wrapper.timeout = timeout
            wrapper.name = name
            return wrapper
        return decorator



from __future__ import absolute_import, unicode_literals

import filecmp
import os
import re
import shutil
import sys
from glob import glob
from io import open

from ksconf.consts import KSCONF_DEBUG, SMART_CREATE, SMART_NOCHANGE, SMART_UPDATE
from ksconf.util.compare import file_compare


def _is_binary_file(filename, peek=2048):
    # https://stackoverflow.com/a/7392391/315892; modified for Python 2.6 compatibility
    textchars = bytearray(set([7, 8, 9, 10, 12, 13, 27]) | set(range(0x20, 0x100)) - set([0x7f]))
    with open(filename, "rb") as f:
        b = f.read(peek)
        return bool(b.translate(None, textchars))


_dir_exists_cache = set()


def dir_exists(directory):
    """ Ensure that the directory exists """
    # This works as long as we never call os.chdir()
    if directory in _dir_exists_cache:
        return
    if not os.path.isdir(directory):
        os.makedirs(directory)
    _dir_exists_cache.add(directory)


def smart_copy(src, dest):
    """ Copy (overwrite) file only if the contents have changed. """
    ret = SMART_CREATE
    if os.path.isfile(dest):
        if file_compare(src, dest):
            # Files already match.  Nothing to do.
            return SMART_NOCHANGE
        else:
            ret = SMART_UPDATE
            os.unlink(dest)
    shutil.copy2(src, dest)
    return ret


def _stdin_iter(stream=None):
    if stream is None:
        stream = sys.stdin
    for line in stream:
        yield line.rstrip()


def file_fingerprint(path, compare_to=None):
    stat = os.stat(path)
    fp = (stat.st_mtime, stat.st_size)
    if compare_to:
        return fp != compare_to
    else:
        return fp


def expand_glob_list(iterable, do_sort=False):
    for item in iterable:
        if "*" in item or "?" in item:
            glob_expanded = glob(item)
            if do_sort:
                glob_expanded.sort()
            for match in glob_expanded:
                yield match
        else:
            yield item


# This is a Splunk-style (props) stanza style glob:
# where '*' is a single path component, and '...' or '**' means any depth
_glob_to_regex = [
    ("**", r".*"),
    ("*", r"[^/\\]*"),
    ("?", r"."),
    ("...", r".*"),
    (".", r"\."),
]
_glob_to_regex_find = "({})".format("|".join(re.escape(r) for r, _ in _glob_to_regex))


def splglob_to_regex(pattern, re_flags=None):
    glob_to_regex = dict(_glob_to_regex)
    regex = re.sub(_glob_to_regex_find, lambda m: glob_to_regex[m.group()], pattern)
    # If NO anchors have been explicitly given, then assume full-match mode:
    if not re.search(r'(?<![\[\\])[$^]', regex):
        regex = "^{}$".format(regex)
    return re.compile(regex, flags=re_flags)


def splglob_simple(pattern):
    """ Return a splglob that either matches a full path or match a simple file """
    if "/" not in pattern:
        # Assume we've been given a simple file name:   app.conf, *.tgz
        pattern = "^.../{}$".format(pattern)
    else:
        pattern = "^{}$".format(pattern)
    return pattern


def relwalk(top, topdown=True, onerror=None, followlinks=False):
    """ Relative path walker
    Like os.walk() except that it doesn't include the "top" prefix in the resulting 'dirpath'.
    """
    top = os.fspath(top)
    if not top.endswith(os.path.sep):
        top += os.path.sep
    prefix = len(top)
    for (dirpath, dirnames, filenames) in os.walk(top, topdown, onerror, followlinks):
        dirpath = dirpath[prefix:]
        yield (dirpath, dirnames, filenames)


def file_hash(path, algorithm="sha256"):
    import hashlib
    h = hashlib.new(algorithm)
    with open(path, "rb") as fp:
        buf = True
        while buf:
            buf = fp.read(4096)
            h.update(buf)
    return h.hexdigest()


def _samefile(file1, file2):
    if hasattr(os.path, "samefile"):
        # Nix
        return os.path.samefile(file1, file2)
    else:
        # Windows
        file1 = os.path.normpath(os.path.normcase(file1))
        file2 = os.path.normpath(os.path.normcase(file2))
        return file1 == file2


class ReluctantWriter:
    """
    Context manager to intelligently handle updates to an existing file.  New content is written
    to a temp file, and then compared to the current file's content.  The file file will be
    overwritten only if the contents changed.
    """

    def __init__(self, path, *args, **kwargs):
        self.path = path
        self._arg = (args, kwargs)
        self._fp = None
        self._tmpfile = path + ".tmp"
        self.change_needed = None
        self.result = None

    def __enter__(self):
        args, kwargs = self._arg
        self._fp = open(self._tmpfile, *args, **kwargs)
        return self._fp

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Don't do anything, other than try to close/delete the file, if an error occurred.
        try:
            self._fp.close()
        except Exception:
            raise
        if exc_type:
            if KSCONF_DEBUG in os.environ:
                # LOG that temp file is being kept
                pass
            else:
                os.unlink(self._tmpfile)
            return
        if not os.path.isfile(self.path):
            os.rename(self._tmpfile, self.path)
            self.change_needed = True
            self.result = "created"
        elif filecmp.cmp(self._tmpfile, self.path):
            os.unlink(self._tmpfile)
            self.change_needed = False
            self.result = "unchanged"
        else:
            os.unlink(self.path)
            os.rename(self._tmpfile, self.path)
            self.change_needed = True
            self.result = "updated"



from __future__ import annotations

import os
import re
from collections import defaultdict
from fnmatch import fnmatch
from os import PathLike
from typing import Type

from ksconf.util.file import relwalk

"""

LayerRootBase has one or more 'Layer', each layer has one or more 'File's.


LayerRoot methods:

    - list_files():   Return superset of all file names returned by all layers (order undefined)
    - walk():         Return list_files() like content in a os.walk() (or relwalker) like way --
                      must consider directory order, useful for copying tree, for example.
                      Assumption for now:  Don't return layer per file, just what files exist.
                      Ask about layers per file later.
    - list_layers():  Iterate over layer objects (metadata retrievable, on demand)
    - get_file():     Return files (in ranked layer order)


Other possible methods:
    list_dirs():      Return list of known directories?   Not sure how we want this part to work.
                      Perhaps walk() is good enough?


dotD style layers:

    LayerRootDotD   has one or more 'LayerMount', each LayerMount has one or more Layer,
                    which has one or more 'File's.



Remember:  This must work for NON-layered directories too, hopefully with minimal overhead.

This must work with an explicitly given list of layers

"""


def _path_join(*parts):
    """ A slightly smarter / more flexible path appender.
    Drop any None or "." elements
    """
    parts = [p for p in parts if p is not None]
    return os.path.join(*parts)


def path_in_layer(layer, path, sep=os.path.sep):
    """ Check to see if path exist within layer.
    Returns either None, or the path without the shared prefix with layer.
    """
    # Using 'sep' over os.path.join / os.path.split should be okay here as we should only ever be
    # given relative paths (no Windows UNC/drive letters)
    if layer is None:
        # Return as-is, since layer is root
        return path
    layer_parts = layer.split(sep)
    layer_count = len(layer_parts)
    path_parts = path.split(sep)
    if len(path_parts) < layer_count:
        return False
    path_suffix = path_parts[:layer_count]
    if layer_parts != path_suffix:
        return False
    return sep.join(path_parts[layer_count:])


# Exceptions

class LayerException(Exception):
    pass


class LayerUsageException(LayerException):
    pass


class LayerFilter:
    _valid_actions = ("include", "exclude")

    def __init__(self):
        self._rules = []

    def add_rule(self, action, pattern):
        # If no filter rules have been setup yet, be sure to set the default
        if action not in self._valid_actions:
            raise ValueError(f"Unknown action of {action}.  "
                             f"Valid actions include: {self._valid_actions}")
        if not self._rules:
            if action == "include":
                first_filter = ("exclude", "*")
            elif "exclude":
                first_filter = ("include", "*")
            self._rules.append(first_filter)
        self._rules.append((action, pattern))

    def evaluate(self, layer: "LayerRootBase.Layer") -> bool:
        response = True
        layer_name = layer.name
        for rule_action, rule_pattern in self._rules:
            if fnmatch(layer_name, rule_pattern):
                response = rule_action == "include"
        return response

    __call__ = evaluate


class LayerConfig:

    def __init__(self):
        # Set defaults
        self.follow_symlink = False
        self.block_files = re.compile(r"\.(bak|swp)$")
        self.block_dirs = {".git"}


class LayerRootBase:
    """ All 'path's here are relative to the ROOT. """

    class File(PathLike):
        __slots__ = ["layer", "relative_path", "_stat"]

        def __init__(self, layer, relative_path, stat=None):
            self.layer = layer
            self.relative_path = relative_path
            self._stat = stat

        def __fspath__(self) -> str:
            return self.physical_path

        @property
        def physical_path(self):
            return _path_join(self.layer.root, self.layer.physical_path, self.relative_path)

        @property
        def logical_path(self):
            return _path_join(self.layer.logical_path, self.relative_path)

        @property
        def stat(self):
            if self._stat is None:
                self._stat = os.stat(self.physical_path)
            return self._stat

        @property
        def size(self):
            return self.stat.st_size

        @property
        def mtime(self):
            return self.stat.st_mtime

    class Layer:
        """ Basic layer Container:   Connects logical and physical paths. """
        __slots__ = ["name", "root", "logical_path", "physical_path", "config", "_file_cls"]

        def __init__(self, name: str, root: str, physical: str, logical: str,
                     config: LayerConfig, file_cls: Type):
            self.name = name
            self.root = root
            self.physical_path = physical
            self.logical_path = logical
            self.config = config
            self._file_cls = file_cls

        def walk(self):
            # In the simple case, this is good enough.   Some subclasses will need to override
            for (root, dirs, files) in relwalk(_path_join(self.root, self.physical_path),
                                               followlinks=self.config.follow_symlink):
                files = [f for f in files if not self.config.block_files.search(f)]
                for d in list(dirs):
                    if d in self.config.block_dirs:
                        dirs.remove(d)
                yield (root, dirs, files)

        def list_files(self):
            File = self._file_cls
            for (top, dirs, files) in self.walk():
                for file in files:
                    yield File(self, _path_join(top, file))

        def get_file(self, path):
            """ Return file object (by logical path), if it exists in this layer. """
            # XXX: There's probably ways to optimize this.  fine for now (correctness over speed)
            File = self._file_cls
            rel_path = path_in_layer(self.logical_path, path)
            if not rel_path:
                return None
            file_ = File(self, rel_path)
            if os.path.isfile(file_.physical_path):
                return file_
            '''
            path_p = _path_join(self.root, self.physical_path, rel_path)
            if os.path.isfile(path_p):
                return File(self, rel_path)
            '''

    def __init__(self, config=None):
        self._layers = []
        self.layer_filter = None
        self.config = config or LayerConfig()

    def apply_filter(self, layer_filter):
        """
        Apply a destructive filter to all layers.  layer_filter(layer) will be called one for each
        layer, if the filter returns True than the layer is kept.  Root layers are always kept.
        """
        layers = [l for l in self._layers if layer_filter(l)]
        result = self._layers != layers
        self._layers = layers
        return result

    def order_layers(self):
        raise NotImplementedError

    def add_layer(self, layer, do_sort=True):
        self._layers.append(layer)
        if do_sort:
            self.order_layers()

    def list_layers(self):
        return self._layers

    def list_layer_names(self):
        return [l.name for l in self.list_layers()]

    def list_files(self):
        """ Return a list of logical paths. """
        files = set()
        for layer in self._layers:
            for file_ in layer.list_files():
                files.add(file_.logical_path)
        return list(files)

    def get_file(self, path):
        """ return all layers associated with the given relative path. """
        for layer in self._layers:
            file_ = layer.get_file(path)
            if file_:
                yield file_

    def get_path_layers(self, path):
        pass


class DirectLayerRoot(LayerRootBase):
    """
    A very simple direct LayerRoot implementation that relies on all layer paths to be explicitly
    given without any automatic detection mechanisms.  You can think of this as the legacy
    implementation.
    """

    def add_layer(self, path):
        Layer, File = self.Layer, self.File
        # Layer name should be considered arbitrary and unimportant here
        layer_name = os.path.basename(path)
        if not os.path.isdir(path):
            raise LayerUsageException("Layers must be directories.  "
                                      f"Given path '{path}' is not a directory.")
        layer = Layer(layer_name, path, None, None, config=self.config, file_cls=File)
        super(DirectLayerRoot, self).add_layer(layer)

    def order_layers(self):
        # No op.  Irrelevant as layers are given (CLI) in the order they should be applied.
        pass


"""


    Layers must be discovered depth-first,
    Files must be walked breath-first?  Is that right?

    MyApp/                          <- Root + MountPointTransparent (contains 1 layer)
        bin/
        default.d/                  <- MountPoint Disconnect from parent layer
            10-upstream             <- Layer
            20-kintyre-core         <- Layer
            99-override             <- Layer
        static/
        lookup.d/



Q:  What about nested layers, should that be supported?  (Yes, this is an horrific example)
A:  Let's support if it just works naturally, otherwise, let's not go to extra lengths to support
A:  Multiple LayerRoots SHOULD be supported.

    MyApp/                          <- LayerRoot & Layer (Anonymous -- lowest ranking)
        lib.d/
            10-upstream
                botocore/
                requests/
            20-kintyre              <-- LayerMount
                botocore.d/
                    10-python3      <-- Layer (nested)
                    20-python2      <-- Layer (nested)
                requests/
        default.d/                  <- Disconnect from parent layer
            10-upstream             <- Layer
            20-kintyre-core         <- Layer
            99-override             <- Layer
        static/

"""


# Q:  How do we mark "mount-points" in the directory structure to keep multiple layers
#     from claiming the same files?????
class DotDLayerRoot(LayerRootBase):

    class Layer(LayerRootBase.Layer):
        __slots__ = ["prune_points"]

        def __init__(self, name, root, physical, logical, config, file_cls, prune_points=None):
            super(DotDLayerRoot.Layer, self).__init__(name, root, physical, logical, config=config,
                                                      file_cls=file_cls)
            self.prune_points = set(prune_points) if prune_points else set()

        def walk(self):
            for (root, dirs, files) in super(DotDLayerRoot.Layer, self).walk():
                if root in self.prune_points:
                    # Cleanup files/dirs to keep walk() from descending deeper
                    del dirs[:]
                else:
                    yield (root, dirs, files)

    '''
    class MountBase:
        def __init__(self, path):
            self.path = path

    class MountTransparent(MountBase):
        """ Pass through files as-is, no manipulation. """
        pass

    class MountDotD(MountBase):
        def __init__(self, path):
            super(DotDLayerRoot.MountDotD, self).__init__(path)
    '''

    mount_regex = re.compile(r"(?P<realname>[\w_.-]+)\.d$")
    layer_regex = re.compile(r"(?P<layer>\d\d-[\w_.-]+)")

    def __init__(self, config=None):
        super(DotDLayerRoot, self).__init__(config)
        # self.root = None
        self._root_layer = None
        self._mount_points = defaultdict(list)

    def apply_filter(self, layer_filter):
        # Apply filter function, but also be sure to keep the root layer
        def fltr(l):
            return l is self._root_layer or layer_filter(l)
        return super(DotDLayerRoot, self).apply_filter(fltr)

    def set_root(self, root, follow_symlinks=None):
        """ Set a root path, and auto discover all '.d' directories.

        Note:  We currently only support '.d/<layer>' directories, a file like
        `default.d/10-props.conf` won't be handled here.
        """
        Layer, File = self.Layer, self.File
        if follow_symlinks is None:
            follow_symlinks = self.config.follow_symlink

        for (top, dirs, files) in relwalk(root, topdown=False, followlinks=follow_symlinks):
            del files

            top_dirname, top_basename = os.path.split(top)
            mount_mo = self.mount_regex.match(top_basename)

            if mount_mo:
                for dir_ in dirs:
                    dir_mo = self.layer_regex.match(dir_)
                    if dir_mo:
                        # XXX: Nested layers breakage, must substitute multiple ".d" folders in `top`
                        layer = Layer(dir_mo.group("layer"),
                                      root,
                                      physical=os.path.join(top, dir_),
                                      logical=os.path.join(top_dirname, mount_mo.group("realname")),
                                      config=self.config,
                                      file_cls=File)
                        self.add_layer(layer)
                        self._mount_points[top].append(dir_)
                    else:
                        # XXX: Give the user the option of logging the near-matches (could indicate a
                        # problem in the config, or could be some other legit directory structure)
                        '''
                        print(f"LAYER NEAR MISS:  {top} looks like a mount point, but {dir_} doesn't "
                              "follow the expected convention")
                        '''
                        pass
            elif top.endswith(".d"):
                '''
                print(f"MOUNT NEAR MISS:  {top}")
                '''
                pass

        # XXX: Adding <root> should be skipped if (and only if) root itself if a '.d' folder
        # Very last operation, add the top directory as the final layer (lowest rank)
        prune_points = [os.path.join(mount, layer) for mount, layers in self._mount_points.items()
                        for layer in layers]
        layer = Layer("<root>", root, None, None, config=self.config, file_cls=File,
                      prune_points=prune_points)
        self.add_layer(layer, do_sort=False)
        self._root_layer = layer

    def list_layers(self):
        # Return all but the root layer.
        # Avoiding self._layers[:-1] because there could be cases where root isn't included.
        return [l for l in self._layers if l is not self._root_layer]

    def order_layers(self):
        # Sort based on layer name (or other sorting priority:  00-<name> to 99-<name>
        self._layers.sort(key=lambda l: l.name)



""" SUBCOMMAND:  ``ksconf check <CONF>``

Usage example:   (Nice pre-commit script)

.. code-block:: sh

    find . -name '*.conf' | ksconf check -

"""
from __future__ import absolute_import, unicode_literals

import os
from collections import Counter

from ksconf.commands import KsconfCmd, dedent
from ksconf.conf.parser import PARSECONF_STRICT_NC, ConfParserException, parse_conf
from ksconf.consts import EXIT_CODE_BAD_CONF_FILE, EXIT_CODE_INTERNAL_ERROR, EXIT_CODE_SUCCESS
from ksconf.util import debug_traceback
from ksconf.util.completers import conf_files_completer
from ksconf.util.file import _stdin_iter, expand_glob_list


class CheckCmd(KsconfCmd):
    help = "Perform basic syntax and sanity checks on .conf files"
    description = dedent("""
    Provides basic syntax and sanity checking for Splunk's .conf
    files.  Use Splunk's built-in ``btool check`` for a more robust
    validation of attributes and values.

    Consider using this utility as part of a pre-commit hook.""")
    maturity = "stable"

    def register_args(self, parser):
        parser.add_argument("conf", metavar="FILE", nargs="+", help=dedent("""\
            One or more configuration files to check.
            If '-' is given, then read a list of files to validate from standard input""")
                            ).completer = conf_files_completer
        parser.add_argument("--quiet", "-q", default=False, action="store_true",
                            help="Reduce the volume of output.")
        ''' # Do we really need this?
        parser.add_argument("--max-errors", metavar="INT", type=int, default=0, help=
            "Abort check if more than this many files fail validation.  "
            "Useful for a pre-commit hook where any failure is unacceptable.")
        '''

    def pre_run(self, args):
        # For Windows users, expand any glob patterns as needed.
        args.conf = list(expand_glob_list(args.conf))

    def run(self, args):
        # Should we read a list of conf files from STDIN?
        if len(args.conf) == 1 and args.conf[0] == "-":
            confs = _stdin_iter()
        else:
            confs = args.conf
        c = Counter()
        exit_code = EXIT_CODE_SUCCESS
        for conf in confs:
            c["checked"] += 1
            if not os.path.isfile(conf):
                self.stderr.write(f"Skipping missing file:  {conf}\n")
                c["missing"] += 1
                continue
            try:
                parse_conf(conf, profile=PARSECONF_STRICT_NC)
                c["okay"] += 1
                if not args.quiet:
                    self.stdout.write(f"Successfully parsed {conf}\n")
                    self.stdout.flush()
            except ConfParserException as e:
                self.stderr.write(f"Error in file {conf}:  {e}\n")
                self.stderr.flush()
                exit_code = EXIT_CODE_BAD_CONF_FILE
                # TODO:  Break out counts by error type/category (there's only a few of them)
                c["error"] += 1
            except Exception as e:  # pragma: no cover
                self.stderr.write(f"Unhandled top-level exception while parsing {conf}.  "
                                  f"Aborting.\n{e}\n")
                debug_traceback()
                exit_code = EXIT_CODE_INTERNAL_ERROR
                c["error"] += 1
                break
        if True:  # show stats or verbose
            self.stdout.write(f"Completed checking {c['checked']} files.  "
                              f"rc={exit_code} Breakdown:\n"
                              f"   {c['okay']} files were parsed successfully.\n"
                              f"   {c['error']} files failed.\n")
        return exit_code



""" SUBCOMMAND:  ``ksconf package -f <SPL> <DIR>``

Usage example:

.. code-block:: sh

   ksconf package -f myapp.tgz MyApp/


Build system example:

.. code-block:: sh

    ksconf package -f release/myapp-{{version}}.tgz \\
            --block-local \\
            --set-version={{git_tag}} \\
            --set-build=${TRAVIS_BUILD_NUMBER:-0}


"""

import argparse
import os

from ksconf.commands import KsconfCmd, dedent
from ksconf.consts import EXIT_CODE_BAD_ARGS, EXIT_CODE_CLI_ARG_DEPRECATED, EXIT_CODE_SUCCESS
from ksconf.package import AppPackager


class PackageCmd(KsconfCmd):
    help = "Create a Splunk app .spl file from a source directory"
    description = dedent("""\
    Create a Splunk app or add on tarball (``.spl``) file from an app directory.

    ``ksconf package`` can do useful things like, exclude unwanted files, combine layers, set the
    application version and build number, drop or promote the ``local`` directory into ``default``.

    Note that some arguments, like the ``FILE`` support special values that can be automatically
    evaluated at runtime.  For example the placeholders ``{{version}}`` or ``{{git_tag}}`` can be
    expanded into the output tarball filename.
    """)
    # format = "manual"
    maturity = "alpha"

    default_blocklist = [
        ".git*",
        "*.py[co]",
        "__pycache__",
        ".DS_Store"
    ]

    def register_args(self, parser: argparse.ArgumentParser):
        def wb_type(action):
            def f(pattern):
                return action, pattern
            return f

        parser.add_argument("source", metavar="SOURCE",
                            help="Source directory for the Splunk app.")
        parser.add_argument("-f", "--file", metavar="SPL",
                            help="Name of splunk app file (tarball) to create.  "
                            "Placeholder variables in ``{{var}}`` syntax can be used here.")

        parser.add_argument("--app-name",
                            help="Specify the top-level app folder name.  "
                                 "If this is not given, the app folder name is automatically "
                                 "extracted from the basename of SOURCE.  "
                                 "Placeholder variables, such as ``{{app_id}}`` can be used here.")
        parser.add_argument("--blocklist", "-b",
                            action="append",
                            default=self.default_blocklist,
                            help="Pattern for files/directories to exclude.  "
                                 "Can be given multiple times.  You can load multiple exclusions "
                                 "from disk by using ``file://path`` which can be used "
                                 "with ``.gitignore`` for example.  (Default includes: {})"
                            .format(", ".join("``{}``".format(i) for i in self.default_blocklist)))

        # XXX: This should be smarter; stacked like we support for layers -- where order matters.
        parser.add_argument("--allowlist", "-a",
                            action="append", default=[],
                            help="Remove a pattern that was previously added to the blocklist.")

        player = parser.add_argument_group(
            "Layer filtering",
            "If the app being packaged includes multiple layers, these arguments can be used to "
            "control which ones should be included in the final app file.  If no layer options "
            "are specified, then all layers will be included.")

        player.add_argument("--layer-method",
                            choices=["dir.d", "disable", "auto"],
                            default="dir.d",
                            help="Set the layer type used by SOURCE.  "
                                 "Additional description provided in in the ``combine`` command."
                                 "Note that 'auto' is no longer supported as of v0.10.")
        player.add_argument("-I", "--include", action="append", default=[], dest="layer_filter",
                            type=wb_type("include"), metavar="PATTERN",
                            help="Name or pattern of layers to include.")
        player.add_argument("-E", "--exclude", action="append", default=[], dest="layer_filter",
                            type=wb_type("exclude"), metavar="PATTERN",
                            help="Name or pattern of layers to exclude from the target.")

        parser.add_argument("--follow-symlink", "-l", action="store_true", default=False,
                            help="Follow symbolic links pointing to directories.  "
                                 "Symlinks to files are always followed.")

        # Set app version or extra app version.
        parser.add_argument("--set-version", metavar="VERSION",
                            help="Set application version.  By default the application version "
                                 "is read from default/app.conf.  Placeholder variables such as "
                                 "``{{git_tag}}`` can be used here.")
        parser.add_argument("--set-build", metavar="BUILD",
                            help="Set application build number.")

        plocal = parser.add_mutually_exclusive_group()
        parser.set_defaults(local="merge")
        plocal.add_argument("--allow-local",
                            dest="local", action="store_const", const="preserve",
                            help="Allow the ``local`` folder to be kept as-is  "
                                 "WARNING: "
                                 "This goes against Splunk packaging practices, and will cause "
                                 "AppInspect to fail.  "
                                 "However, this option can be useful for private package transfers "
                                 "between servers, app backups, or other admin-like tasks.")
        plocal.add_argument("--block-local",
                            dest="local", action="store_const", const="block",
                            help="Block the ``local`` folder and ``local.meta`` from the package.")
        plocal.add_argument("--merge-local",
                            dest="local", action="store_const", const="merge",
                            help="Merge any files in ``local`` into the ``default`` folder during "
                                 "packaging.  This is the default behavior.")

        pbuild = parser.add_argument_group("Advanced Build Options",
                                           "The following options are for more advanced app "
                                           "building workflows.")
        pbuild.add_argument("--release-file",
                            help="Write the path of the newly generated archive file (SPL) after "
                                 "the archive is written.  "
                                 "This is useful in build scripts when the SPL contains variables "
                                 "so the final name may not be known ahead of time.")

    @staticmethod
    def load_blocklist(path):
        with open(path) as stream:
            for line in stream:
                line = line.rstrip()
                if line and not line.startswith("#"):
                    yield line

    def pre_run(self, args):
        if "local" in args.blocklist:
            self.stderr.write("Blocking 'local' is not supported.   "
                              "Most likely you want '--block-local' instead.\n")
            return EXIT_CODE_BAD_ARGS

        blocklist = []
        for pattern in args.blocklist:
            if pattern.startswith("file://"):
                blocklist_file = pattern[7:]
                expanded = list(self.load_blocklist(blocklist_file))
                self.stderr.write("Extended blocklist from {} with {:d} entries\n"
                                  .format(blocklist_file, len(expanded)))
                blocklist.extend(expanded)
            else:
                blocklist.append(pattern)
        args.blocklist = [pattern for pattern in blocklist if pattern not in args.allowlist]

    def run(self, args):
        ''' Create a Splunk app/add-on .spl file from a directory '''

        # Just call combine (writing to a temporary directory) and the tar it up.  At some point
        # we should do it all in memory, but for now this good enough.  For more sophisticated
        # builds use a temp build directory.  This is what ksconf's BuildManager does.

        app_name = args.app_name
        app_name_source = "set via commandline"
        if not app_name:
            app_name = os.path.basename(args.source)
            app_name_source = "taken from source directory"
            '''
            if os.path.basename(args.source) == ".":
                app_name = os.path.basename(os.getcwd())
                app_name_source = "extracted from working directory"
            else:
                app_name = os.path.basename(args.source)
                app_name_source = "extracted from source directory"
        '''
        self.stdout.write(f"Packaging {app_name}   (App name {app_name_source})\n")
        packager = AppPackager(args.source, app_name, output=self.stderr)

        if args.layer_method == "auto":
            # There'd no way to make this option legal but not shown in argparse.  :-(
            # Yeah, all this needs *LOTS* of work!
            self.stderr("The 'auto' option for layer_method is not longer supported.  "
                        "This will be an error in v0.11 and removed in v0.12\n")
            from ksconf import __version__
            if __version__.startswith("0.11."):
                return EXIT_CODE_CLI_ARG_DEPRECATED

        # XXX:  Make the combine step optional.  Either via detection (no .d folders/layers) OR manually opt-out
        #       for faster packaging in simple scenarios (this may not matter once this is done in memory)
        with packager:
            packager.combine(args.source, args.layer_filter,
                             layer_method=args.layer_method,
                             allow_symlink=args.follow_symlink)
            # Handle local files
            if args.local == "merge":
                packager.merge_local()
            elif args.local == "block":
                packager.block_local()
            elif args.local == "preserve":
                pass
            else:   # pragma: no cover
                raise ValueError(f"Unknown value for 'local': {args.local}")

            if args.blocklist:
                self.stderr.write(f"Applying blocklist:  {args.blocklist!r}\n")
                packager.blocklist(args.blocklist)

            if args.set_build or args.set_version:
                packager.update_app_conf(
                    version=args.set_version,
                    build=args.set_build)

            packager.check()
            # os.system(f"ls -lR {packager.app_dir}")

            dest = args.file or "{}-{{{{version}}}}.tgz".format(packager.app_name.lower().replace("-", "_"))
            archive_path = packager.make_archive(dest)
            self.stderr.write("Archive created:  file={} size={:.2f}Kb\n".format(
                os.path.basename(archive_path), os.stat(archive_path).st_size / 1024.0))

            if args.release_file:
                # Should this be expanded to be an absolute path?
                with open(args.release_file, "w") as f:
                    f.write(archive_path)

        return EXIT_CODE_SUCCESS



""" Splunk Application facts:

Easily collect Splunk app name, version, label, and other nuggets from ``app.conf``

"""

from __future__ import absolute_import, annotations, unicode_literals

from collections import defaultdict
from dataclasses import asdict, dataclass, field, fields
from io import StringIO
from os import fspath
from pathlib import Path
from typing import ClassVar

from ksconf.app.manifest import AppArchiveContentError
from ksconf.archive import extract_archive, gaf_filter_name_like
from ksconf.compat import List, Tuple
from ksconf.conf.merge import merge_conf_dicts
from ksconf.conf.parser import (PARSECONF_LOOSE, ConfType, conf_attr_boolean,
                                default_encoding, parse_conf)


@dataclass
class AppFacts:
    """ Basic Splunk application info container.
    A majority of these facts are extracted from ``app.conf``
    """
    name: str
    label: str = None
    id: str = None
    version: str = None
    author: str = None
    description: str = None
    state: str = None
    build: int = None

    is_configured: bool = field(init=False, default=None)
    allows_disable: bool = field(init=False, default=None)
    state_change_requires_restart: bool = field(init=False, default=None)

    install_source_checksum: str = field(init=False, default=None)
    install_source_local_checksum: str = field(init=False, default=None)
    check_for_updates: bool = field(init=False, default=None)
    is_visible: bool = field(init=False, default=None)

    deployer_lookups_push_mode: str = field(init=False, default=None)
    deployer_push_mode: str = field(init=False, default=None)

    _conf_translate_pairs: ClassVar[List[Tuple[str, List[str]]]] = [
        ("launcher", [
            "version",
            "author",
            "description"]),
        ("install", [
            "state",
            "build",
            "is_configured",
            "allows_disable",
            "install_source_checksum",
            "install_source_local_checksum",
            "state_change_requires_restart",
        ]),
        ("package", [
            "id",
            "check_for_updates"]),
        ("ui", [
            "label",
            "is_visible"]),
        ("shclustering", [
            "deployer_lookups_push_mode",
            "deployer_push_mode"]
         ),
    ]

    def to_dict(self) -> dict:
        return asdict(self)

    def to_tiny_dict(self, *keep_attrs) -> dict:
        """ Return dict representation, discarding the Nones """
        return {k: v for k, v in asdict(self).items() if v is not None or k in keep_attrs}

    @classmethod
    def from_conf(cls, name, conf: ConfType) -> "AppFacts":
        """
        Create AppFacts from an app.conf configuration content.
        """
        new = cls(name)

        type_mapping = {f.name: f.type for f in fields(cls) if f.type in ("bool", "int")}

        def convert_attr(attr_name, value):
            # XXX: Is there a better approach for dataclasses?  fields() returns a list
            data_type = type_mapping.get(attr_name, None)
            convert_function = None
            if data_type == "bool":
                convert_function = conf_attr_boolean
            elif data_type == "int":
                convert_function = int
            else:
                return value
            try:
                return convert_function(value)
            except ValueError:
                # Best effort.  pass the buck to the consumer
                return value

        for stanza_name, attributes in cls._conf_translate_pairs:
            if stanza_name in conf:
                stanza = conf[stanza_name]
                for attr in attributes:
                    if attr in stanza:
                        setattr(new, attr, convert_attr(attr, stanza[attr]))
        return new

    @classmethod
    def from_app_dir(cls, app_path: Path) -> "AppFacts":
        """
        Create an AppFacts from a local file system.  This expects a standard
        (non-layered) installed or extracted app folder.  Both ``default`` and
        ``local`` are considered.
        """
        app_path = Path(app_path)
        app_conf_paths = [
            app_path / "default" / "app.conf",
            app_path / "local" / "app.conf"
        ]
        conf = {}
        for app_conf_path in app_conf_paths:
            if app_conf_path.is_file():
                conf = merge_conf_dicts(conf, parse_conf(app_conf_path, PARSECONF_LOOSE))
        return cls.from_conf(app_path.name, conf)

    @classmethod
    def from_archive(cls, archive: Path):
        ''' Returns list of app names, merged app_conf and a dictionary of extra facts that may be useful '''
        archive = Path(archive)

        app_names = set()
        app_confs = defaultdict(dict)

        is_app_conf = gaf_filter_name_like("app.conf")
        for gaf in extract_archive(archive, extract_filter=is_app_conf):
            app_name, relpath = gaf.path.split("/", 1)
            if relpath.endswith("/app.conf") and gaf.payload:
                conf_folder = relpath.rsplit("/")[0]
                tmp = StringIO(gaf.payload.decode(default_encoding))
                tmp.name = fspath(archive.joinpath(gaf.path))
                app_confs[conf_folder] = parse_conf(tmp, profile=PARSECONF_LOOSE)
                del tmp
            app_names.add(app_name)
            del app_name, relpath

        if len(app_names) > 1:
            raise AppArchiveContentError(
                f"Found multiple top-level app names!  Archive {archive} "
                f"contains apps {', '.join(app_names)}")

        # Merge default and local (in the correct order)
        conf = merge_conf_dicts(app_confs["default"], app_confs["local"])
        app_name = app_names.pop()
        return cls.from_conf(app_name, conf)



from __future__ import absolute_import, unicode_literals


def build_rest_namespace(base, owner=None, app=None):
    if owner and app:
        p = (base, "servicesNS", owner, app)
    elif app:
        p = (base, "servicesNS", "nobody", app)
    elif owner:
        raise ValueError("Can't specify user without app!")
    else:
        p = (base, "services")
    return "/".join(p)


def build_rest_url(base, service, owner=None, app=None):
    prefix = build_rest_namespace(base, owner, app)
    return prefix + "/" + service



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK
""" KSCONF - Ksconf Splunk CONFig tool

Optionally supports argcomplete for commandline argument (tab) completion.

Install & register with:

     pip install argcomplete
     activate-global-python-argcomplete  (in ~/.bashrc)

"""

from __future__ import absolute_import, unicode_literals

import argparse
import os
import platform
import sys
from collections import defaultdict

import ksconf
import ksconf.util
from ksconf.commands import DescriptionHelpFormatterPreserveLayout, get_all_ksconf_cmds
from ksconf.consts import EXIT_CODE_ENV_BUSTED, EXIT_CODE_INTERNAL_ERROR, KSCONF_DEBUG
from ksconf.util.completers import autocomplete

# Workaround PY2:  WindowsError: [Error -2146893795] Provider DLL failed to initialize correctly
# Someday need to re-evaluate this and see if it's reproducible on other machines....
try:
    from random import choice
except OSError:  # WindowsError:  pragma: no cover
    def choice(options):
        return options[0]


###################################################################################################
# CLI definition


# ------------------------------------------ wrap to 80 chars ----------------v
_cli_description = """KSCONF: Ksconf Splunk CONFig tool

This utility handles a number of common Splunk app maintenance tasks in a small
and easy to deploy package.  Specifically, this tool deals with many of the
nuances with storing Splunk apps in git and pointing live Splunk apps to a git
repository.  Merging changes from the live system's (local) folder to the
version controlled (default) folder and dealing with more than one layer of
"default" are all supported tasks which are not native to Splunk.
"""
# ------------------------------------------ wrap to 80 chars ----------------^


def check_py_sane():
    """ Run a simple python environment sanity check.   Here's the scenario, if Splunk's
    python is called but not all the correct environment variables have been set, then ksconf can
    fail in unclear ways.
    """
    try:
        # There must be a more 'sane' way to check this.  But for now, this test works...
        from hashlib import md5
        del md5
    except ImportError:
        return False
    return True


def handle_cmd_failed(subparser, ep):
    """ Build a bogus subparser for a cmd that can't be loaded, with the only purpose of providing
    a more consistent user experience. """
    # Not sure how much *better* this is.  But it at least it gets away from the dumb stares
    # when the subcommand silently disappears.  (Visible from ksconf --version, but still...
    # It's confusing, even if *just* during development)
    marker = "*" * 80
    description = f"{marker}\n***   {ep.error}\n{marker}"
    badparser = subparser.add_parser(ep.name, description=description,
                                     help=f"****** {ep.error} ******",
                                     formatter_class=DescriptionHelpFormatterPreserveLayout)

    def handler(args):
        sys.stderr.write("Unable to process due to internal error in "
                         f"{ep.name}'\n{ep.error}\n")
        return EXIT_CODE_INTERNAL_ERROR
    badparser.set_defaults(funct=handler)
    # Consume all remaining args.  But if params are passed first sometimes the user still sees
    # "unrecognized arguments".  A deeper hack is needed to improve beyond this.
    badparser.add_argument('args', nargs=argparse.REMAINDER)


def build_cli_parser(do_formatter=False):
    parser_kwargs = dict(
        fromfile_prefix_chars="@",
        description=_cli_description,
        prog="ksconf")
    if sys.version_info > (3, 5):
        # Disable abbreviations as they could lead to accidental assignment as the CLI grows.
        parser_kwargs["allow_abbrev"] = False
    if do_formatter:
        parser_kwargs["formatter_class"] = DescriptionHelpFormatterPreserveLayout
    parser = argparse.ArgumentParser(**parser_kwargs)
    subparsers = parser.add_subparsers()

    # XXX: Lazyload version information; this launches 'git' which is expensive.
    version_info = []

    # XXX:  Check terminal size before picking a signature
    version_info.append(choice(ksconf.__ascii_sigs__))

    verbuild = f"%(prog)s {ksconf.__version__}"
    if ksconf.__build__:
        verbuild += f"  (Build {ksconf.__build__})"
    version_info.append(verbuild)
    version_info.append(f"Python: {sys.version.split()[0]}  ({sys.executable})")
    if ksconf.__vcs_info__:
        version_info.append(ksconf.__vcs_info__)
    install_location = os.path.dirname(os.path.abspath(ksconf.__file__))
    version_info.append(f"Installed at: {install_location}")
    version_info.append(f"Platform:  {platform.version()}")
    try:
        from ksconf.vc.git import git_version
        git_ver = git_version()
        if git_ver:
            version_info.append(f"Git support:  ({git_ver['path']}) {git_ver['version']}")
        else:
            version_info.append("Git support:  'git' not found in PATH")
    except Exception as e:
        # Shouldn't happen, but we really don't blowup!
        version_info.append(f"Git support:  Detection failed!  {e}")
    # XXX:  Grab splunk version and home, if running as a splunk app
    version_info.append(f"Written by {ksconf.__author__}.")
    version_info.append(f"Copyright {ksconf.__copyright__}, all rights reserved.")
    version_info.append(f"Licensed under {ksconf.__license__}")

    # Add entry-point subcommands
    subcommands = defaultdict(list)

    # XXX:  Eventually lazy load subcommands to save resources.   (Low priority)
    for ep in get_all_ksconf_cmds(on_error="return"):
        # (name, entry, cmd_cls, error)
        dist = ep.entry.dist
        distro = ""
        if hasattr(dist, "version"):
            if hasattr(dist, "name"):
                # entrypoints (required by ksconf)
                distro = f"{dist.name}  ({ep.entry.dist.version})"
            elif hasattr(dist, "location") and hasattr(dist, "project_name"):   # pragma: no cover
                # Attributes per pkg_resource  (currently disabled)
                distro = f"{dist.project_name}  ({dist.version})  @{dist.location}"

        subcommands[distro].append((ep.name, ep.cmd_cls, ep.error))
        if ep.cmd_cls:
            cmd = ep.cmd_cls(ep.entry.name)
            # XXX: Find a better way to handle argparse errors: (TypeError) ex: invalid arguments
            cmd.add_parser(subparsers)
        elif ep.error:
            handle_cmd_failed(subparsers, ep)

    for distro_name, items in sorted(subcommands.items()):
        if distro_name:
            version_info.append(f"\n  {distro_name}\n\n    Commands:")
        else:
            version_info.append("\n\n    Commands:")
        for (name, cmd_cls, error) in items:
            if cmd_cls is None:
                m = "(?)"
            else:
                m = f"({cmd_cls.maturity})"
            if error:
                version_info.append(f"      {name:15} {m:10}  {error}")
            else:
                info = f"      {name:15} {m:10}  OK"
                if cmd_cls.version_extra:
                    info = f"{info}   ({cmd_cls.version_extra})"
                version_info.append(info)

    # Common settings
    '''
    ### DEPRECATE THESE
    parser.add_argument("-S", "--duplicate-stanza", default=DUP_EXCEPTION, metavar="MODE",
                        choices=[DUP_MERGE, DUP_OVERWRITE, DUP_EXCEPTION],
                        help="Set duplicate stanza handling mode.  If [stanza] exists more than "
                             "once in a single .conf file:  Mode 'overwrite' will keep the last "
                             "stanza found.  Mode 'merge' will merge keys from across all stanzas, "
                             "keeping the the value form the latest key.  Mode 'exception' "
                             "(default) will abort if duplicate stanzas are found.")
    parser.add_argument("-K", "--duplicate-key", default=DUP_EXCEPTION, metavar="MODE",
                        choices=[DUP_EXCEPTION, DUP_OVERWRITE],
                        help="Set duplicate key handling mode.  A duplicate key is a condition "
                             "that occurs when the same key (key=value) is set within the same "
                             "stanza.  Mode of 'overwrite' silently ignore duplicate keys, "
                             "keeping the latest.  Mode 'exception', the default, aborts if "
                             "duplicate keys are found.")
    '''
    parser.add_argument('--version', action='version', version="\n".join(version_info))
    parser.add_argument("--force-color", dest="tty_mode", const="force", action="store_const",
                        default=os.environ.get("KSCONF_TTY_COLOR", "auto"),
                        help="Force TTY color mode on.  Useful if piping the output a color-aware "
                             "pager, like 'less -R'")
    parser.add_argument("--disable-color", dest="tty_mode", const="off", action="store_const",
                        help="Disable TTY color mode.  This can also be setup as environmental variable: "
                        "``export KSCONF_TTY_COLOR=off``")

    # Logging settings -- not really necessary for simple things like 'diff', 'merge', and 'sort';
    # more useful for 'patch', very important for 'combine'

    return parser


def cli(argv=None, _unittest=False):
    # TODO:  Rename '_unitest' to something more appropriate, maybe _exit=True?
    check_py()

    parser = build_cli_parser(True)
    if not _unittest:
        autocomplete(parser)

    args = parser.parse_args(argv)

    ksconf.util.terminal.TTY_COLOR_MODE = args.tty_mode

    if not hasattr(args, "funct") or args.funct is None:
        sys.stderr.write(parser.format_usage())
        sys.exit(1)

    try:
        return_code = args.funct(args)
    except Exception as e:  # pragma: no cover
        # Set KSCONF_DEBUG=1 to enable a traceback
        if _unittest:
            from ksconf.consts import KSCONF_DEBUG
            os.environ[KSCONF_DEBUG] = "1"
        sys.stderr.write(f"Unhandled top-level exception ({type(e).__name__}):  {e}\n")
        ksconf.util.debug_traceback()
        return_code = EXIT_CODE_INTERNAL_ERROR

    if _unittest:
        return return_code or 0
    else:  # pragma: no cover
        sys.exit(return_code or 0)


def check_py():
    if not check_py_sane():
        # TODO: This should ALSO check to make sure this is a Splunk-based install, or this 'help' will be misguided.
        sys.stderr.write("Doh!  Environmental configuration issue found preventing 'ksconf' from running.\n")
        # TODO:  We should show the Windows equivalent, but primarily this is an issue on Linux.
        # TODO:  If we're install as a splunk app, we should be able to give the real path to SPLUNK_HOME,
        #        which quite likely is ALSO not set.
        sys.stderr.write("\n\n")  # Often there's crap on the console from warnings.  Whitespace!
        sys.stderr.write("Try running this command first:  source $SPLUNK_HOME/bin/setSplunkEnv\n")
        # Allow   `KSCONF_DEBUG=1 ksconf --version` to run, even if environmental issues exist
        if KSCONF_DEBUG not in os.environ:
            sys.exit(EXIT_CODE_ENV_BUSTED)
    # Okay, now NEVER call this code again....   (helpful for unit-testing & nested calls)
    globals()["check_py"] = lambda: None



# -*- coding: utf-8 -*-
""" Splunk App helper classes

Note that these representations are for native Splunk apps that use 'default'
and 'local' and have not built-in concept of ksconf layers.
"""

from __future__ import absolute_import, annotations, unicode_literals

from pathlib import Path

from ksconf.app.facts import AppFacts
from ksconf.app.manifest import AppManifest
from ksconf.compat import Tuple


def get_facts_manifest_from_archive(
        archive: Path,
        calculate_hash=True,
        check_paths=True) -> Tuple[AppFacts, AppManifest]:
    """ Get both AppFacts and AppManifest from a single archive.
    If ``calculate_hash`` is True, then the manifest will contain checksums for
    all files in the archive.  Without this, it's not possible to calculate a
    hash for the combined manifest.

    Use this function to collect both metadata about the app and a full listing
    of the app's contents.
    """
    # XXX: Optimize to create AppFacts and AppManifest concurrently; from a single read of the archive.
    # XXX: Use this in ksconf.commands.unarchive
    archive = Path(archive)

    facts = AppFacts.from_archive(archive)
    manifest = AppManifest.from_archive(archive, calculate_hash=calculate_hash)
    if check_paths:
        manifest.check_paths()

    return facts, manifest



from __future__ import unicode_literals

import re
import sys
from enum import Enum


class SmartEnum(Enum):
    CREATE = "created"
    UPDATE = "updated"
    NOCHANGE = "unchanged"

    def __str__(self):
        return self.value


class _UNSET:
    pass


UNSET = _UNSET()


MANIFEST_HASH = "sha256"


# Legacy names
SMART_CREATE = SmartEnum.CREATE
SMART_UPDATE = SmartEnum.UPDATE
SMART_NOCHANGE = SmartEnum.NOCHANGE


# EXIT_CODE_* constants:  Use consistent exit codes for scriptability
#
#   0-9    Normal/successful conditions
#   20-49  Error conditions (user caused)
#   50-59  Externally caused (should retry)
#   100+   Internal error (developer required)
#   120+   Environmental error

EXIT_CODE_SUCCESS = 0
EXIT_CODE_NOTHING_TO_DO = 1
EXIT_CODE_USER_QUIT = 2
EXIT_CODE_NO_SUCH_FILE = 5
EXIT_CODE_MISSING_ARG = 6
EXIT_CODE_BAD_ARGS = 7

EXIT_CODE_DIFF_EQUAL = 0
EXIT_CODE_DIFF_CHANGE = 3
EXIT_CODE_DIFF_NO_COMMON = 4
EXIT_CODE_FORMAT_APPLIED = 8
EXIT_CODE_SORT_APPLIED = 9

EXIT_CODE_CLI_ARG_DEPRECATED = 10

# Errors caused by users
EXIT_CODE_BAD_CONF_FILE = 20
EXIT_CODE_BAD_ARCHIVE_FILE = 21
EXIT_CODE_FAILED_SAFETY_CHECK = 22
EXIT_CODE_COMBINE_MARKER_MISSING = 30

# Errors caused by GIT interactions
EXIT_CODE_GIT_FAILURE = 40

# Retry or temporary failure
EXIT_CODE_EXTERNAL_FILE_EDIT = 50

# Unresolvable issues (developer required)
EXIT_CODE_INTERNAL_ERROR = 100
EXIT_CODE_FEAT_NOT_IMPLEMENTED = 101     # Too bad we can't use 404 :=)

# Environmental error
EXIT_CODE_ENV_BUSTED = 120
EXIT_CODE_BAD_PY_VERSION = 121


# This gets properly supported in Python 3.6, but until then....
RegexType = type(re.compile(r'.'))


KSCONF_DEBUG = "KSCONF_DEBUG"


# Cleanup namespace for wildcard imports
del re, sys



from __future__ import absolute_import, unicode_literals

import fnmatch
import os
import re
import sys
from collections import Counter

from ksconf.conf.parser import GLOBAL_STANZA
from ksconf.util.file import splglob_to_regex

# It seems like each item on the list should be capability of having it's own type and flags (which could be inherited, at may not be known at the time the rules are first initialized)
# The should still be some phase that each "Rule" goes though (1) creation to set core attributes, (2) prep where things can be compiled, syntax checked, and any default flags should be available by this time, and (3) evalute against a specific value for true/false.
# This would allow things like a special prefix to lazy-switch modes (for example:  ~regex~, or only do fnmatching if there's a wildcard otherwise stick with simple string matching, ...)


class FilteredList:
    IGNORECASE = 1
    INVERT = 2
    VERBOSE = 4

    def __init__(self, flags=0, default=True):
        self.data = []
        self.rules = None
        self.counter = Counter()
        self.flags = flags
        self._prep = True
        #  If no patterns defined, return default.  (True => match everything)
        self.default = default

    def _feed_from_file(self, path):
        items = []
        with open(path) as f:
            for line in f:
                line = line.rstrip()
                # Skip empty or "comment" lines
                if line and line[0] != "#":
                    items.append(line)
        if self.flags & self.VERBOSE:
            sys.stderr.write(f"Loaded {len(items)} patterns from {path}\n")
        return items

    def feed(self, item, filter=None):
        if item.startswith("file://"):
            # File ingestion mode
            filename = item[7:]
            for item in self._feed_from_file(filename):
                self.feed(item, filter)
        else:
            if filter:
                item = filter(item)
            self.data.append(item)
        # New items added.  Mark prep-work as incomplete
        self._prep = False

    def feedall(self, iterable, filter=None):
        if iterable:
            for i in iterable:
                self.feed(i, filter)
        return self

    def _pre_match(self):  # pragma: no cover
        pass

    def match(self, item):
        if self.data:
            # Kick off any first-time preparatory activities
            if self._prep is False:
                self._pre_match()
                self.reset_counters()
                self._prep = True

            ret = self._match(item)
            if ret:
                self.counter[ret] += 1
                result = True
            else:
                result = False

        else:
            #  No patterns defined, use default
            return self.default
        if self.flags & self.INVERT:
            return not result
        else:
            return result

    def match_path(self, path):
        if os.path.sep != "/":
            path = path.replace(os.path.sep, "/")
        return self.match(path)

    def match_stanza(self, stanza):
        """ Same as match(), but handle GLOBAL_STANZA gracefully. """
        if stanza is GLOBAL_STANZA:
            stanza = "default"
        return self.match(stanza)

    def reset_counters(self):
        # Set all the counters to 0, so the caller can know which filters had 0 hits
        self.counter = Counter()
        self.counter.update((n, 0) for n in self.data)

    @property
    def has_rules(self):
        return bool(self.data)

    def _match(self, item):  # pragma: no cover
        """ Return name of rule, indicating a match or not. """
        raise NotImplementedError


class FilteredListString(FilteredList):
    """ Handle simple string comparisons """

    def _pre_match(self):
        if self.flags & self.IGNORECASE:
            # Lower-case all strings in self.data.  (Only need to do this once)
            self.rules = {i.lower() for i in self.data}
        else:
            self.rules = set(self.data)
        return self.rules

    def _match(self, item):
        if self.flags & self.IGNORECASE:
            item = item.lower()
        if item in self.rules:
            return item
        else:
            return False

    def reset_counters(self):
        self.counter = Counter()
        self.counter.update({n: 0 for n in self.rules})


class FilteredListRegex(FilteredList):
    """ Regular Expression support """

    def calc_regex_flags(self):
        re_flags = 0
        if self.flags & self.IGNORECASE:
            re_flags |= re.IGNORECASE
        return re_flags

    def _pre_match(self):
        # Compile all regular expressions
        re_flags = self.calc_regex_flags()
        # XXX: Add better error handling here for friendlier user feedback
        self.rules = [(pattern, re.compile(pattern, re_flags)) for pattern in self.data]

    def _match(self, item):
        for name, pattern_re in self.rules:
            if pattern_re.match(item):
                return name
        return False

    def reset_counters(self):
        self.counter = Counter()
        self.counter.update({i[0]: 0 for i in self.rules})


class FilteredListWildcard(FilteredListRegex):
    """ Wildcard support (handling '*' and ?')
    Technically fnmatch also supports [] and [!] character ranges, but we don't advertise that
    """

    def _pre_match(self):
        # Use fnmatch to translate wildcard expression to a regex, and compile regex
        re_flags = self.calc_regex_flags()
        self.rules = [(wc, re.compile(fnmatch.translate(wc), re_flags)) for wc in self.data]


class FilteredListSplunkGlob(FilteredListRegex):
    """ Classic wildcard support ('*' and ?') plus '...' or '**' for multiple-path components with
    some (non-advertised) pass-through regex behavior
    """

    def _pre_match(self):
        # Use splglob_to_regex to translate wildcard expression to a regex, and compile regex
        re_flags = self.calc_regex_flags()
        self.rules = [(wc, splglob_to_regex(wc, re_flags)) for wc in self.data]


class_mapping = {
    "string": FilteredListString,
    "wildcard": FilteredListWildcard,
    "regex": FilteredListRegex,
    "splunk": FilteredListSplunkGlob,
}


def create_filtered_list(match_mode, flags=0, default=True):
    try:
        class_ = class_mapping[match_mode]
    except KeyError:
        raise NotImplementedError(f"Matching mode {match_mode!r} undefined")
    return class_(flags, default)



# -*- coding: utf-8 -*-
"""
"""

from __future__ import absolute_import, annotations, unicode_literals

from collections import Counter
from dataclasses import asdict, dataclass, field, fields
from enum import Enum
from os import fspath
from pathlib import Path, PurePath, PurePosixPath
from typing import Set

from ksconf.app import AppManifest
from ksconf.archive import extract_archive
from ksconf.compat import List
from ksconf.util.compare import cmp_sets

# Deployment Action classes


class DeployActionType(Enum):
    SET_APP_NAME = "app"
    SOURCE_REFERENCE = "source"
    EXTRACT_FILE = "extract_file"
    REMOVE_FILE = "remove"

    """ Implement in future phase
    SET_SYMLINK = "link"
    UPDATE_META = "meta"
    """

    def __str__(self):
        return self.value


# Specific action definitions

@dataclass
class DeployAction:
    # Abstract base class
    action: str

    def to_dict(self) -> dict:
        data = asdict(self)
        data["action"] = str(self.action)

        for dc_field in fields(self):
            if dc_field.name == "action":
                continue

            value = data[dc_field.name]
            t = eval(dc_field.type)
            if issubclass(t, PurePath):
                value = fspath(value)
                data[dc_field.name] = value
        return data

    @classmethod
    def from_dict(self, data: dict) -> "DeployAction":
        data = data.copy()
        action = data.pop("action")
        action = DeployActionType(action)
        da_class = get_deploy_action_class(action)
        for dc_field in fields(da_class):
            if dc_field.name == "action":
                continue
            value = data[dc_field.name]
            t = eval(dc_field.type)
            if issubclass(t, PurePath):
                value = t(value)
                data[dc_field.name] = value
        return da_class(**data)


@dataclass
class DeployAction_ExtractFile(DeployAction):
    action: str = field(init=False, default=DeployActionType.EXTRACT_FILE)
    subtype: str
    path: PurePosixPath
    mode: int = None
    mtime: int = None
    hash: str = None
    rel_path: str = None


@dataclass
class DeployAction_RemoveFile(DeployAction):
    action: str = field(init=False, default=DeployActionType.REMOVE_FILE)
    path: PurePosixPath


@dataclass
class DeployAction_SetAppName(DeployAction):
    action: str = field(init=False, default=DeployActionType.SET_APP_NAME)
    name: str


@dataclass
class DeployAction_SourceReference(DeployAction):
    action: str = field(init=False, default=DeployActionType.SOURCE_REFERENCE)
    archive_path: str
    hash: str


def get_deploy_action_class(action: str) -> DeployAction:
    return {
        DeployActionType.EXTRACT_FILE: DeployAction_ExtractFile,
        DeployActionType.REMOVE_FILE: DeployAction_RemoveFile,
        DeployActionType.SET_APP_NAME: DeployAction_SetAppName,
        DeployActionType.SOURCE_REFERENCE: DeployAction_SourceReference,
    }[action]


'''
@dataclass
class DeployIntention:
    """ Container describing a deployment transformation intention. """
    path: Path
    action: DeployIntentionActions
    detail: Any = None

    _detail_class_mapping: ClassVar = {
        DeployIntentionActions.EXTRACT_FILE: _DeployAction_Extract,
        DeployIntentionActions.SET_APP_NAME: _DeployAction_SetAppName,
    }

    def __post_init__(self):
        detail_cls = self._detail_class_mapping.get(self.action)
        if detail_cls:
            if isinstance(self.detail, dict) and self.detail:
                self.detail = detail_cls(**self.detail)
        self.detail = detail_cls()
'''


class DeploySequence:
    def __init__(self):
        self.actions: List[DeployAction] = []
        self.actions_by_type = Counter()

    def add(self, action: str, *args, **kwargs):
        if not isinstance(action, DeployAction):
            action_cls = get_deploy_action_class(action)
            action = action_cls(*args, **kwargs)
        else:
            assert not args and not kwargs
        self.actions_by_type[action.action] += 1
        self.actions.append(action)

    def to_dict(self) -> dict:
        d = {
            "actions": [a.to_dict() for a in self.actions]
        }
        return d

    @classmethod
    def from_dict(cls, data: dict) -> "DeploySequence":
        seq = cls()
        for action in data["actions"]:
            da = DeployAction.from_dict(action)
            seq.add(da)
        return seq

    @classmethod
    def from_manifest(
            cls,
            manifest: AppManifest) -> "DeploySequence":
        """
        Fresh deploy of an app from scratch.

        (There should probably be a new
        op-code for this, eventually instead of listing every single file.)
        """
        dc = cls()
        dc.add(DeployAction_SourceReference(manifest.source, manifest.hash))
        dc.add(DeployAction_SetAppName(manifest.name))
        for f in manifest.files:
            dc.add(DeployActionType.EXTRACT_FILE, "create", f.path, mode=f.mode, hash=f.hash)
        return dc

    @classmethod
    def from_manifest_transformation(
            cls,
            base: AppManifest,
            target: AppManifest) -> "DeploySequence":
        seq = cls()
        if target is None:
            return cls.from_manifest(base)
        if base.name != target.name:
            raise ValueError(f"Manifest must have the same app name.  {base.name} != {target.name}")
        seq.add(DeployAction_SourceReference(target.source, target.hash))
        seq.add(DeployAction_SetAppName(target.name))

        base_files = {f.path: f for f in base.files}
        target_files = {f.path: f for f in target.files}

        base_only, common, target_only = cmp_sets(base_files, target_files)

        for fn in target_only:
            f = target_files[fn]
            seq.add(DeployAction_ExtractFile("create", fn, mode=f.mode, hash=f.hash))

        for fn in common:
            base_file = base_files[fn]
            target_file = target_files[fn]
            if base_file != target_file:
                sub = "attr" if base_file.content_match(target_file) else "update"
                seq.add(DeployAction_ExtractFile(sub, fn, target_file.mode, hash=target_file.hash))

        for fn in base_only:
            seq.add(DeployAction_RemoveFile(fn))

        return seq


# Will we need this, or can we get all this as class methods within DeploySequence?
class DeployPlanner():
    pass


def _path_by_part_len(path: Path):
    # sort helper function for directory top-down/bottom-up operations
    return len(path.parts), path


class DeployApply:
    def __init__(self, dest: Path):
        self.dest = dest
        self.dir_mode = 0o770

    def resolve_source(self, source, hash):
        # In the future, this may look in a local/remote directory based on the hash value.
        return Path(source)

    def apply_sequence(self, deployment_sequence: DeploySequence):
        '''
        Apply a pre-calculated deployment sequence to the local file system.
        '''
        #
        '''
        app = None
        items = []
        for action in deployment_sequence.actions:
            if action.action == DeployActionType.SET_APP_NAME:
                if items:
                    self.apply_sequence_for_app(archive, deployment_sequence)
                app = action.name
        '''

        '''
    def apply_sequence_for_app(self,
            archive: Path,
            app_name: str,
            deployment_sequence: DeploySequence):
        '''
        keep_paths: Set[Path] = set()
        make_dirs: Set[Path] = set()
        remove_path: Set[Path] = set()
        app_path = Path()
        for action in deployment_sequence.actions:
            if isinstance(action, DeployAction_ExtractFile):
                path = app_path.joinpath(action.path)
                keep_paths.add(path)
                make_dirs.add(path.parent)
            elif isinstance(action, DeployAction_SetAppName):
                app_path = Path(action.name)
            elif isinstance(action, DeployAction_RemoveFile):
                remove_path.add(app_path.joinpath(action.path))
            elif isinstance(action, DeployAction_SourceReference):
                archive = self.resolve_source(action.archive_path, action.hash)
            else:
                raise TypeError(f"Unable to handle action of type {type(action)}")

        # Make necessary directories
        for d in sorted(make_dirs, key=_path_by_part_len):
            dest_dir: Path = self.dest.joinpath(d)
            # Even with careful sorting, we must still use parents=True as some
            # directories contain no files, such as 'ui' in 'default/data/ui/nav'.
            dest_dir.mkdir(self.dir_mode, parents=True, exist_ok=True)

        # Expand matching files
        for gaf in extract_archive(archive):
            p = Path(gaf.path)
            if p in keep_paths:
                dest_path: Path = self.dest.joinpath(p)
                dest_path.write_bytes(gaf.payload)
                # Should we use the chmod from the archive, or the one from the deployment sequence object?
                dest_path.chmod(gaf.mode)

        # Cleanup removed files
        for p in remove_path:
            full_path: Path = self.dest.joinpath(p)
            if full_path.is_file():
                full_path.unlink()

        # Cleanup any empty directories (longest paths first)
        for d in sorted(set(f.parent for f in remove_path),
                        key=_path_by_part_len,
                        reverse=True):
            full_path = self.dest.joinpath(d)
            if full_path.stat().st_nlink == 2:
                # print(f"Cleaning empty directory {d}")
                try:
                    full_path.rmdir()
                except IOError:
                    pass


def expand_archive_by_manifest(
        archive: Path,
        dest: Path,
        manifest: AppManifest,
        dir_mode=0o770):
    """
    Expand an tarball to a local file system including only the files referenced
    by the files within the app manifest.

    This function assumes that safety checks on manifest have already been
    performed, such as eliminating any absolute paths.
    """
    # XXX: Optimize out inefficiencies created by our use of extract_archive()

    ''' If no ``manifest`` is provided, all files are expanded.
    Not sure we want to allow this.  Let's keep this disabled unless needed.
    if manifest is None:
        # It's assumed that this is non-normal code path; as this is inefficient
        manifest = AppManifest.from_archive(archive, calculate_hash=False)
    '''
    keep_paths = set()
    make_dirs = set()
    app_path = Path(manifest.name)
    for f in manifest.files:
        path = app_path.joinpath(f.path)
        keep_paths.add(path)
        make_dirs.add(path.parent)

    # Ensure shorter paths are created first
    make_dirs = [(len(d.parts), d) for d in make_dirs]

    # Make necessary directories
    for _, d in sorted(make_dirs):
        dest_dir: Path = dest.joinpath(d)
        dest_dir.mkdir(dir_mode, parents=True, exist_ok=True)

    # Expand matching files
    for gaf in extract_archive(archive):
        p = Path(gaf.path)
        if p in keep_paths:
            dest_path: Path = dest.joinpath(p)
            dest_path.write_bytes(gaf.payload)
            dest_path.chmod(gaf.mode)
    # Anything else?



from __future__ import unicode_literals


def fileobj_compare(f1, f2):
    # Borrowed from filecmp
    f1.seek(0)
    f2.seek(0)
    buffsize = 8192
    while True:
        b1 = f1.read(buffsize)
        b2 = f2.read(buffsize)
        if b1 != b2:
            return False
        if not b1:
            return True


def file_compare(fn1, fn2):
    with open(fn1, "rb") as f1, \
            open(fn2, "rb") as f2:
        return fileobj_compare(f1, f2)


def cmp_sets(a, b):
    """ Result tuples in format (a-only, common, b-only) """
    set_a = set(a)
    set_b = set(b)
    a_only = sorted(set_a.difference(set_b))
    common = sorted(set_a.intersection(set_b))
    b_only = sorted(set_b.difference(set_a))
    return (a_only, common, b_only)


# For backwards compability
_cmp_sets = cmp_sets



# -*- coding: utf-8 -*-
"""
SUBCOMMAND:  ``ksconf rest-export --output=script.sh <CONF>``

Usage example:

.. code-block:: sh

    ksconf rest-export --output=apply_props.sh /opt/splunk/etc/app/Splunk_TA_aws/local/props.conf


NOTE:

    If we add support for Windows CURL, then we'll need to also support proper quoting for the '%'
    character.   This can be done with '%^', wonky, I know...

"""
from __future__ import absolute_import, unicode_literals

import os
import shlex
import sys
from argparse import ArgumentParser, FileType
from urllib.parse import quote

from ksconf.commands import ConfFileType, KsconfCmd, dedent
from ksconf.conf.parser import GLOBAL_STANZA, PARSECONF_LOOSE
from ksconf.consts import EXIT_CODE_SUCCESS
from ksconf.util.completers import conf_files_completer
from ksconf.util.rest import build_rest_url


class Literal:
    def __init__(self, value):
        self.value = value

    '''
    def __str__(self):
        return self.value
    '''


class CurlCommand:
    def __init__(self):
        self.url = None
        self.method = None  # curl defaults this to POST
        self.pre_args = ["-k"]
        self.post_args = []
        self.headers = {}
        self.data = {}
        self.pretty_format = True

    @classmethod
    def quote(cls, s):
        if isinstance(s, Literal):
            return s.value
        if "$" in s:
            s = f'"{s}"'
        elif " " in s:
            s = f"'{s}'"
        return s

    def get_command(self):
        cmd = ["curl"]
        args = []

        if self.method:
            args.append("-X")
            args.append(self.method)

        if self.headers:
            for header in self.headers:
                value = self.headers[header]
                args.append("-H")
                args.append(f"{header}: {value}")
        if self.data:
            for key in self.data:
                value = self.data[key]
                if self.pretty_format:
                    args.append(Literal("\\\n -d"))
                else:
                    args.append("-d")
                args.append(f"{quote(key)}={quote(value)}")

        # if self.pre_args:
        #    cmd.append(" ".join(self.pre_args))
        cmd.extend(self.pre_args)
        cmd.append(self.url)
        cmd.extend(self.quote(arg) for arg in args)
        cmd.extend(self.quote(arg) for arg in self.post_args)
        return " ".join(cmd)

    def extend_args(self, args):
        # Use shlex parsing to handle embedded quotess
        for s in shlex.split(args):
            self.post_args.append(s)


class RestExportCmd(KsconfCmd):
    help = "Export .conf settings as a curl script to apply to a Splunk instance later (via REST)"
    description = dedent("""\
    Build an executable script of the stanzas in a configuration file that can be later applied to
    a running Splunk instance via the Splunkd REST endpoint.

    This can be helpful when pushing complex props and transforms to an instance where you only have
    UI access and can't directly publish an app.

    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser: ArgumentParser):
        parser.add_argument("conf", metavar="CONF", nargs="+",
                            type=ConfFileType("r", "load", parse_profile=PARSECONF_LOOSE),
                            help="Configuration file(s) to export settings from."
                            ).completer = conf_files_completer
        parser.add_argument("--output", "-t", metavar="FILE",
                            type=FileType("w"), default=sys.stdout,
                            help="Save the shell script output to this file.  "
                                 "If not provided, the output is written to standard output.")

        prsout = parser.add_argument_group("Output Control")

        '''
        prsout.add_argument("--syntax", choices=["curl", "powershell"],  # curl-windows?
                            default="curl",
                            help="Pick the output syntax mode.  "
                                 "Currently only 'curl' is supported.")
        '''
        prsout.add_argument("--disable-auth-output", action="store_true", default=False,
                            help="Turn off sample login curl commands from the output.")
        prsout.add_argument("--pretty-print", "-p", action="store_true", default=False,
                            help=dedent("""\
            Enable pretty-printing.
            Make shell output a bit more readable by splitting entries across lines."""))

        parsg1 = parser.add_mutually_exclusive_group(required=False)
        parsg1.add_argument("-u", "--update", action="store_true", default=False,
                            help="Assume that the REST entities already exist.  "
                                 "By default, output assumes stanzas are being created.")
        parsg1.add_argument("-D", "--delete", action="store_true", default=False,
                            help=dedent("""\
            Remove existing REST entities.  This is a destructive operation.
            In this mode, stanza attributes are unnecessary and ignored.
            NOTE:  This works for 'local' entities only; the default folder cannot be updated.
            """))

        parser.add_argument("--url", default="https://localhost:8089",
                            help="URL of Splunkd.  Default:  %(default)s")
        parser.add_argument("--app", default="$SPLUNK_APP",
                            help="Set the namespace (app name) for the endpoint")

        parser.add_argument("--user", help="Deprecated.  Use --owner instead.")
        parser.add_argument("--owner", default="nobody",
                            help="Set the object owner.  Typically, the default of 'nobody' is "
                                 "ideal if you want to share the configurations at the app-level.")
        parser.add_argument("--conf", dest="conf_type", metavar="TYPE",
                            help=dedent("""\
            Explicitly set the configuration file type.  By default, this is derived from CONF, but
            sometimes it's helpful to set this explicitly.  Can be any valid Splunk conf file type.
            Examples include: 'app', 'props', 'tags', 'savedsearches', etc."""))

        parser.add_argument("--extra-args", action="append",
                            help=dedent("""\
            Extra arguments to pass to all CURL commands.
            Quote arguments on the command line to prevent confusion between arguments to ksconf vs
            curl."""))

    @staticmethod
    def build_rest_url(base, owner, app, conf):
        # XXX: Quote owner & app; however for now we're still allowing the user to pass though an
        #  environmental variable as-is and quoting would break that.   Need to make a decision,
        # for now this is not likely to be a big issue given app and user name restrictions.
        return build_rest_url(base, f"configs/conf-{conf}", owner, app)

    def run(self, args):
        ''' Convert a conf file into a bunch of CURL commands'''
        r"""

        Some inspiration in the form of CURL commands...

        [single_quote_kv]
        REGEX = ([^=\s]+)='([^']+)'
        FORMAT = $1::$2
        MV_ADD = 0

        CREATE NEW:

        curl -k https://SPLUNK:8089/servicesNS/nobody/my_app/configs/conf-transforms \
         -H "Authorization: Splunk $SPLUNKDAUTH" -X POST \
         -d name=single_quote_kv \
         -d REGEX="(%5B%5E%3D%5Cs%5D%2B)%3D%27(%5B%5E%27%5D%2B)%27" \
         -d FORMAT='$1::$2'

        UPDATE EXISTING:  (note the change in URL/name attribute)

        curl -k https://SPLUNK:8089/servicesNS/nobody/my_app/configs/conf-transforms/single_quote_kv \
         -H "Authorization: Splunk $SPLUNKDAUTH" -X POST \
         -d REGEX="(%5B%5E%3D%5Cs%5D%2B)%3D%27(%5B%5E%27%5D%2B)%27" \
         -d FORMAT='$1::$2' \
         -d MV_ADD=0
        """
        stream = args.output

        if args.user:       # pragma: no cover
            from warnings import warn
            warn("Use '--owner' instead of '--user'", DeprecationWarning)
            if args.owner != "nobody":
                raise ValueError("Can't use both --user and --owner at the same time!")
            args.owner = args.user

        if args.pretty_print:
            line_breaks = 2
        else:
            line_breaks = 1

        if args.disable_auth_output is False:
            # Make this preamble optional
            stream.write("## Example of creating a local SPLUNKDAUTH token\n")
            stream.write("export SPLUNKDAUTH=$("
                         f"curl -ks {args.url}/services/auth/login -d username=admin -d password=changeme "
                         "| grep sessionKey "
                         r"| sed -E 's/[ ]*<sessionKey>(.*)<.sessionKey>/\1/')")
            stream.write('; [[ -n $SPLUNKDAUTH ]] && echo "Login token created"')
            stream.write("\n\n\n")

        for conf_proxy in args.conf:
            conf = conf_proxy.data
            if args.conf_type:
                conf_type = args.conf_type
            else:
                conf_type = os.path.basename(conf_proxy.name).replace(".conf", "")

            stream.write(f"# CURL REST commands for {conf_proxy.name}\n")

            for stanza_name, stanza_data in conf.items():
                cc = CurlCommand()
                cc.pretty_format = args.pretty_print
                cc.url = self.build_rest_url(args.url, args.owner, args.app, conf_type)
                if args.extra_args:
                    for extra_arg in args.extra_args:
                        cc.extend_args(extra_arg)

                if stanza_name is GLOBAL_STANZA:
                    # XXX:  Research proper handling of default/global stanzas..
                    # As-is, curl returns an HTTP error, but yet the new entry is added to the
                    # conf file.  So I suppose we could ignore the exit code?!    ¯\_(ツ)_/¯
                    stream.write("### WARN:  Writing to the default stanza may not work as "
                                 "expected.  Or it may work, but be reported as a failure.  "
                                 "Patches welcome!\n")
                    cc.url += "/default"
                elif args.update or args.delete:
                    cc.url += "/" + quote(stanza_name, "")  # Must quote '/'s too.
                else:
                    cc.data["name"] = stanza_name

                if args.delete:
                    cc.method = "DELETE"
                else:
                    # Add individual keys
                    for (key, value) in stanza_data.items():
                        cc.data[key] = value

                cc.headers["Authorization"] = "Splunk $SPLUNKDAUTH"

                stream.write(cc.get_command())
                stream.write("\n" * line_breaks)
            stream.write("\n\n" * line_breaks)
        stream.write("\n")

        return EXIT_CODE_SUCCESS



""" ksconf.builder.steps:  Collection of reusable build steps for reuse in your build script.
"""

import re
import sys
from shutil import copy2, rmtree
from typing import List

from ksconf.builder import QUIET, VERBOSE, BuildStep


def clean_build(step: BuildStep) -> None:
    """ Ensure a clean build folder for consistent build results."""
    log = step.get_logger()
    if step.build_path.is_dir():
        log("Purging previous build folder")
        rmtree(step.build_path)
    else:
        log("Make build folder")
    step.build_path.mkdir()

    if not step.dist_path.is_dir():
        log("Made dist dir")
        step.dist_path.mkdir()


def copy_files(step: BuildStep,
               patterns: List[str],
               target: str = None) -> None:
    """ Copy source files into the build folder that match given glob patterns """
    log = step.get_logger()
    if target:
        log(f"Copying files into build folder under {target}")
    else:
        log("Copying files into build folder")
    log(f"Copy src={step.source_path} to build={step.build_path} target={target}", VERBOSE * 2)
    dirs = files = 0
    for pattern in patterns:
        if pattern.endswith("/"):
            log(f"Looking for all files under '{pattern}'", VERBOSE * 3)
            pattern += "**/*"
        elif "*" in pattern:
            log(f"Looking for all files matching '{pattern}'", VERBOSE * 3)
        else:
            log(f"Looking for files named '{pattern}'", VERBOSE * 3)
        file_per_pattern = 0
        for f in step.source_path.glob(pattern):
            relative = f.relative_to(step.source_path)
            if target:
                dest = step.build_path / target / relative
            else:
                dest = step.build_path / relative

            dest_parent = dest.parent
            if not dest_parent.is_dir():
                log(f"Mkdir {dest_parent}", VERBOSE)
                dest_parent.mkdir(parents=True)
                dirs += 1
            if f.is_file():
                log(f"Copy  {f}", VERBOSE)
                copy2(f, dest)
                files += 1
                file_per_pattern += 1
        log(f"Copied {file_per_pattern} files matching '{pattern}'", VERBOSE * 2)
    # TODO: Expand capabilities to capture files/dirs per pattern, helpful to get lookup counts
    log(f"Completed copying {len(patterns)} patterns.  Created {files} files in {dirs} directories")


def _get_python_info_rename(path: str) -> str:
    """ Figure out the name of the package, without the version for renaming the directory.
    Why do all this?  Well, (1) it's tricky to figure the boundary between package name and version
    for some packages, and this approach removes the guesswork, (2) sometimes you want to distribute
    this metadata about the upstream packages bundled with a Splunk app, but you don't want to leave
    a mess behind after app upgrades.  Or, sometime packages need access to entrypoints stored
    within these folders.   But whenever possible, just delete 'em.
    """
    if path.name.endswith(".egg-info"):
        f = "PKG-INFO"
    else:
        # Assume dist-info.  Are there other options?
        f = "METADATA"
    pkgmetainfodata = path / f
    with pkgmetainfodata.open() as f:
        for line in f:
            match = re.match(r'^Name: ([A-Z-a-z].+)', line)
            if match:
                name = match.group(1)
                break
            if not line.strip():
                # First blank line; gone too far; give up
                return
        else:
            return
        return name + path.suffix


def pip_install(step: BuildStep,
                requirements_file: str = "requirements.txt",
                dest: str = "lib",
                python_path: str = None,
                isolated: bool = True,
                dependencies: bool = True,
                handle_dist_info: str = "remove",  # or 'rename'
                remove_console_scripts: bool = True
                ) -> None:
    dist_info_options = ("remove", "rename", "keep")
    if handle_dist_info not in dist_info_options:
        raise ValueError(f"Expecting 'handle_dist_info' to be one of {dist_info_options}")

    log = step.get_logger()
    if python_path is None:
        python_path = sys.executable
    target = step.build_path / dest

    extra_args = []

    if isolated:
        extra_args.append("--isolated")
    if not dependencies:
        extra_args.append("--no-deps")

    step.run(python_path, "-m", "pip",          # '-m pip' is reliable; avoids pip/pip2/pip3
             "install",                         # Install mode (no upgrade needed; fresh install)
             "-r", requirements_file,           # File to read packages from
             "--target", target,                # Targeted (directory) mode
             "--disable-pip-version-check",     # Warnings aren't helpful here
             "--no-compile",                    # Avoid creating *.pyc files
             *extra_args)

    log("pip installation completed successfully", QUIET)

    #  With the "--no-compile" options, this shouldn't be needed.  Keeping for now.
    for unwanted in target.rglob("*.py[co]"):
        log(f"Remove unwanted {unwanted}", VERBOSE * 2)
        unwanted.unlink()

    # Remove any console-script entry points files (bin|Script); depends on package
    if remove_console_scripts:
        for folder in ("bin", "Scripts"):
            path = target / folder
            if path.is_dir():
                log(f"Removing console-script folder: {path}")
                rmtree(path)

    if handle_dist_info == "keep":
        return

    log(f"Handling {{dist,egg}}-info folders:  mode={handle_dist_info} folder={target}", VERBOSE)
    di_handled = 0
    for path in target.iterdir():
        if path.is_dir() and path.suffix in (".dist-info", ".egg-info"):
            if handle_dist_info == "remove":
                log(f"Remove unwanted dist-info folder: {path.name}", VERBOSE * 2)
                rmtree(path)
                di_handled += 1
            elif handle_dist_info == "rename":
                try:
                    new_name = _get_python_info_rename(path)

                    if new_name:
                        log(f"Replacing dist-info {path} with {new_name}", VERBOSE * 2)
                        path.rename(path.with_name(new_name))
                        di_handled += 1
                    else:
                        log("Dist-info rename failed.  Unable to determine package name for "
                            f"{path}", QUIET)
                except Exception as e:
                    log(f"Exception during Dist-info rename for {path}:  {e}", QUIET * 2)
    if di_handled:
        log(f"Dist info:  {handle_dist_info}d {di_handled} directories")
